""" Base controller class for superwise package """
import json
import traceback
from abc import ABC
from abc import ABCMeta
from abc import abstractproperty
from urllib.parse import urlencode

from requests import Response

from superwise import logger
from superwise.models.data_entity import DataEntity
from superwise.models.dataset import Dataset
from superwise.models.model import Model
from superwise.models.notification import Notification
from superwise.models.project import Project
from superwise.models.role import Role
from superwise.models.segment import Segment
from superwise.models.transaction import Transaction
from superwise.models.validation_error import ValidationError
from superwise.models.version import Version
from superwise.utils.client import Client
from superwise.utils.exceptions import *


class BaseController(ABC):
    """Base class for controllers"""

    __metaclass__ = ABCMeta

    def __init__(self, client: Client, sw):
        """

        ### Args:

        `client`: an instance of client object

        sw`: superwise object


        """
        self.client = client
        self.path = None
        self.model_name = abstractproperty()
        self.logger = logger
        self.model = None
        self.sw = sw

    def _dict_to_model(self, params, model_name=None):
        model_name = model_name or self.model_name
        try:
            if "self" in params:
                del params["self"]
            if isinstance(params, list):
                model = list()
                for param in params:
                    cmodel = globals()[model_name](**param)
                    model.append(cmodel)
            else:
                if "__class__" in params:
                    del params["__class__"]
                model = globals()[model_name](**params)
        except Exception as err:
            traceback.print_exc()
            raise Exception("exception in create {}".format(err))
        return model

    def post(self):
        """
        ### Description:

        Prepare data and call run API POST call

        ### Return:
        requests response object
        """
        params = self.model.get_properties()
        self.logger.info("POST %s ", self.path)
        response = self.client.post(self.client.build_url(self.path), params)
        return response

    def patch(self, path=None, params=None) -> Response:
        """
        ### Description:
        Prepare data and call run API PATCH call

        ### Return:

        requests response object
        """
        path = path or self.path
        params = params or self.model.get_properties()
        self.logger.info("PATCH %s ", self.path)
        res = self.client.patch(self.client.build_url(path), params)
        return res

    def update(self, idx, params=None) -> Response:
        """
        ### Description:
        Prepare data and call run API PATCH call using a dictionary as params

        ### Return:

        requests response object
        """
        url = self.client.build_url("{}/{}".format(self.path, idx))
        params = params or self.model.get_properties()
        self.logger.info("PATCH %s ", url)
        res = self.client.patch(url, params)
        return res

    # TODO: Fix response type description in this function
    def parse_response(self, response, model_name=None, is_return_model=True):
        """
        ### Description:

        Format the response, change it from dict to be model based

        ### Args:

        `response`: response object (created by requests package)


        `model_name`: model name (dataentity, version etc)

        `is_return_model`: return model if True or response.body if False

        """
        model_name = model_name or self.model_name
        try:
            body = json.loads(response.content)
        except Exception as ex:
            raise Exception("error loading json, status code {} text {}".format(response.status_code, response.content))
        res = self._dict_to_model(body, model_name=model_name) if is_return_model else body
        return res

    def _create_update(self, model, is_return_model=True, create=True, model_name=None, **kwargs):
        """
        ### Description:

        create/update object from an instance of model

        ### Args:

        `model`: model name (dataentity, version etc)

        `is_return_model`: return model if True or response.body if False

        `create`: create if true, otherwise update


        """

        action = "create" if create else "update"
        try:
            if model.__class__.__name__ == self.model_name:
                self.model = model
                res = self.post() if create else self.patch()
            else:
                raise Exception(
                    "Model {} passed instead of {} to {}".format(
                        model.__class__.__name__, self.model_name, self.__class__.__name__
                    )
                )
            return self.parse_response(res, is_return_model=is_return_model)
        except SuperwiseValidationException:
            raise
        except Exception as e:
            traceback.print_exc()
            raise Exception("exception in {} action: {}".format(action, e))

    def create(self, model, return_model=True, **kwargs):
        """
        ### Description:

        Create object from an instance of model

        ### Args:

        `model`: model name (dataentity, version etc)

        `is_return_model`: return model if True or response.body if False

        """
        return self._create_update(model, create=True, return_model=return_model, **kwargs)

    def get_by_name(self, name):
        """ "
        ### Description:

        Get entities by name, a syntactic sugar for get with name as filter

        ### Args:

        `name`: a string represent the name to filter by

        ### Return:

        a list of entities
        """
        return self.get({"name": name})

    def get(self, fields):
        """
        ### Description:

        wrapper for requests.get()

        ### Args:

        `fields`:  optinal fields for filtering

        ### Return:

        parsed data (dict) from backend
        """
        uri = self.client.build_url(self.path)

        fields_formatted = urlencode(fields)
        uri = "{}?{}".format(uri, fields_formatted)
        self.logger.info("GET with fields %s ", uri)
        res = self.client.get(uri)
        return self.parse_response(res)

    def get_by_id(self, idx, timeout=None):
        """
        ### Description:

        Get a specific resource by id

        ### Args:
        `idx`:  id of entity to fetch

        ### Return:

        parsed model data (dict) from backend
        """

        url = self.client.build_url("{}/{}".format(self.path, idx))
        self.logger.info("GET  %s ", url)
        res = self.client.get(url, timeout=(timeout, timeout))
        return self.parse_response(res)

    def delete(self, model):
        """
        ### Description:

        Delete an object

        ### Args:

        `model`:  a Superwise.Models object

        ### Return:

        parsed response from server (dict)
        """
        if not model.id:
            raise SuperwiseException("Delete allowed only to model exist in server, this is probably a new model")
        return self.delete_by_id(model.id)

    def delete_by_id(self, idx):
        """
        ### Description:

        Delete an object by a given id

        ### Args:

        `idx`:  id to delete

        ### Return:

        parsed response from server (dict)
        """
        url = self.client.build_url("{}/{}".format(self.path, idx))
        self.logger.info("DELETE  %s ", url)
        res = self.client.delete(url)
        return res

    def __str__(self):
        properties = self.model.get_properties()
        return f"<Superwise.{self.__class__.__name__} " f"{properties}"



""" This module implement tasks functionality  """
from superwise.controller.base import BaseController


class ProjectController(BaseController):
    """Class ProjectController - responsible for task functionality"""

    def __init__(self, client, sw):
        """

        ### Args:

        `client`: superwise client object

        `sw`: superwise  object

        """
        super().__init__(client, sw)
        self.path = "model/v1/projects"
        self.model_name = "Project"



import pytest
import requests
from requests import Response

from superwise.models.project import Project


@pytest.fixture(scope="function")
def mock_project_requests(monkeypatch):
    get_response = Response()
    get_response._content = b'{"id": 1, "name": "super cool project", "description": "here we have the most amazing project", "created_at": "2022-05-02T16:10:45.318683", "created_by": "me"}'
    get_response.status_code = 201

    delete_response = Response()
    delete_response._content = b"1"
    delete_response.status_code = 201

    monkeypatch.setattr(requests, "post", lambda *args, **kwargs: get_response)
    monkeypatch.setattr(requests, "get", lambda *args, **kwargs: get_response)
    monkeypatch.setattr(requests, "delete", lambda *args, **kwargs: delete_response)


@pytest.fixture(scope="function")
def mock_multi_project_requests(monkeypatch):
    get_response = Response()
    get_response._content = b'[{"id": 1, "name": "super cool project", "description": "here we have the most amazing project", "created_at": "2022-05-02T16:10:45.318683", "created_by": "me"},{"id": 1, "name": "super cool project", "description": "here we have the most amazing project", "created_at": "2022-05-02T16:10:45.318683", "created_by": "me"} ]'
    get_response.status_code = 201

    monkeypatch.setattr(requests, "get", lambda *args, **kwargs: get_response)


def test_create_project(mock_project_requests, sw):
    p = Project(name="super cool project", description="here we have the most amazing project")
    response = sw.project.create(p)
    assert response.id == 1 and response.name == p.name


def test_delete_project(mock_project_requests, sw):
    project = Project(id=1, name="super cool project", description="here we have the most amazing project")
    res = sw.project.delete(project)
    assert res.status_code == 201
    assert res.json() in [project.id, f"{project.id}"]


def test_get_project_by_id(mock_project_requests, sw):
    project = Project(id=1, name="super cool project", description="here we have the most amazing project")
    ret_project = sw.project.get_by_id(project.id)
    assert isinstance(ret_project, Project)
    assert project.name == ret_project.name


def test_get_project_by_name(mock_multi_project_requests, sw):
    project = Project(id=1, name="super cool project", description="here we have the most amazing project")
    projects = sw.project.get_by_name(project.name)
    assert len(projects) == 2



import os
from enum import Enum


def get_bool(key, default):
    """
    ### Description:

    Cast string to bool

    ### Args:

    `key`: the key to get from ENV variable

    `default`: value to return if key not in ENV variable

    ### Return:
    True/False (bool)
    """

    v = os.environ.get(key, default)
    if v in ["False", "0", False]:
        return False
    else:
        return True


class DefaultValues(Enum):
    SUPERWISE_DOMAIN = "superwise.ai"
    AUTH_URL = "https://auth.superwise.ai"
    ENVIRONMENT = "production"
    SUPERWISE_HOST = "portal.superwise.ai"
    MANAGED_AUTH_URL = "https://auth.managed.superwise.ai"


class Config:
    """
    ### Description:

    Configuration class - define some variables used by package
    """

    ENVIRONMENT = os.environ.get("ENVIRONMENT", DefaultValues.ENVIRONMENT.value)
    SUPERWISE_HOST = os.environ.get("SUPERWISE_HOST", DefaultValues.SUPERWISE_HOST.value)
    POOLING_INTERVAL_SEC = 15
    LIST_DROP_DATA_COLS = ["model_id", "version_id"]
    FILE_SIZE_LIMIT_BYTES = 100000000
    FILE_SIZE_PADDING = 50000000
    INTERNAL_BUCKET_SETTINGS = None
    AUTH_URL = DefaultValues.AUTH_URL.value



from superwise.utils.client import Client
from superwise.utils.storage.internal_storage.aws import AWSInternalStorage
from superwise.utils.storage.internal_storage.azure import AzureInternalStorage
from superwise.utils.storage.internal_storage.gcs import GCSInternalStorage
from superwise.utils.storage.internal_storage.internal_storage import CloudProvider
from superwise.utils.storage.internal_storage.internal_storage import InternalStorage


class InternalStorageFactory:
    @staticmethod
    def get_internal_bucket(superwise_web_client: Client, settings: dict) -> InternalStorage:
        cloud_provider = settings.get("cloud_provider")
        if cloud_provider == CloudProvider.AZURE.value:
            return AzureInternalStorage(
                superwise_web_client, settings.get("azure_account_name"), settings.get("azure_container_name")
            )
        if cloud_provider == CloudProvider.GCP.value:
            return GCSInternalStorage(superwise_web_client, settings.get("gcs_bucket_name"))
        if cloud_provider == CloudProvider.AWS.value:
            return AWSInternalStorage(superwise_web_client, settings.get("s3_bucket_name"))
        else:
            raise ValueError(f"cloud provider {cloud_provider} isn't recognized")



""" This module implement dataset functionality  """
import os
import time
import urllib.parse
from typing import List
from typing import Union

from requests import ReadTimeout

from superwise.controller.base import BaseController
from superwise.models.dataset import Dataset
from superwise.resources.superwise_enums import DatasetOnFailureOptions
from superwise.resources.superwise_enums import DatasetStatus
from superwise.utils.exceptions import SuperwiseDatasetFailureError
from superwise.utils.exceptions import SuperwiseTimeoutError
from superwise.utils.storage import validate_files_size
from superwise.utils.storage.azure import AzureClientReader
from superwise.utils.storage.gcs import GcsClient
from superwise.utils.storage.local import load_from_local_storage
from superwise.utils.storage.s3 import S3Client


class DatasetController(BaseController):
    """Datasets controller class, implement functionalities for dataset API"""

    def __init__(self, client, sw, internal_bucket):
        """

        ### Args:

        `client`: superwise client object

        `sw`: superwise object

        """
        super().__init__(client, sw)
        self.path = "model/v1/datasets"
        self.model_name = "Dataset"
        self._internal_bucket = internal_bucket
        self._scheme_to_func = {
            "gs": self._get_file_from_gcs,
            "s3": self._get_file_from_s3,
            "abfs": self._get_file_from_blob_storage,
        }

    @staticmethod
    def _get_file_from_gcs(file_path, service_account, validate_size, **kwargs) -> dict:
        gcs = GcsClient()
        return gcs.get_file_from_gcs(file_path=file_path, service_account=service_account, validate_size=validate_size)

    @staticmethod
    def _get_file_from_s3(file_path, access_key_id, secret_access_key, role_arn, validate_size, **kwargs) -> dict:
        s3_client = S3Client()
        return s3_client.get_s3_file(
            file_path=file_path,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            role_arn=role_arn,
            validate_size=validate_size,
        )

    @staticmethod
    def _get_file_from_blob_storage(file_path, connection_string, validate_size, **kwargs) -> dict:
        azure_client = AzureClientReader(connection_string=connection_string)
        return azure_client.get_file_from_blob_storage(file_path=file_path, validate_size=validate_size)

    def _upload_files_to_storage(
        self,
        file_paths: Union[str, List[str]],
        project_id: int,
        gcs_service_account: str = None,
        aws_access_key_id: str = None,
        aws_secret_access_key: str = None,
        aws_role_arn: str = None,
        azure_connection_string: str = None,
    ):
        auth_params = dict(
            service_account=gcs_service_account,
            access_key_id=aws_access_key_id,
            secret_access_key=aws_secret_access_key,
            role_arn=aws_role_arn,
            connection_string=azure_connection_string,
        )
        total_size = 0
        for i, file_path in enumerate(file_paths):
            external_path = file_path
            parsed_url = urllib.parse.urlparse(file_path)
            storage_scheme = parsed_url.scheme

            if storage_scheme in self._scheme_to_func:
                get_file_func = self._scheme_to_func[storage_scheme]
                file = get_file_func(file_path, validate_size=True, **auth_params)
            else:
                file = load_from_local_storage(file_path, validate_size=True)
                external_path = f"file://{file['filename']}"
            total_size += file["size"]
            validate_files_size(filenames=file_paths[: i + 1], size=total_size)

            base_name, extension = os.path.splitext(file["filename"])
            file_name = self._internal_bucket.generate_file_name(prefix=base_name, extension=extension)
            blob_path = f"datasets/project_id={project_id}/{file_name}"
            inner_path = self._internal_bucket.upload(blob_path=blob_path, content=file["data"])
            yield inner_path, external_path

    @staticmethod
    def _extract_inner_paths(file_paths: List[tuple]) -> List[str]:
        return [file_path[0] for file_path in file_paths]

    @staticmethod
    def _extract_external_paths(file_paths: List[tuple]) -> List[str]:
        return [file_path[1] for file_path in file_paths]

    def _get_field(self, dataset_id, field):
        dataset = self.get_by_id(idx=dataset_id)

        return getattr(dataset, field)

    def _get_timeout_func(self, dataset_id, dataset_statuses: List[DatasetStatus], timeout_seconds=30):
        _started_at = time.perf_counter()

        def __wait_to_status():
            dataset = self.get_by_id(dataset_id)
            status = dataset.status
            _time_since_start = time.perf_counter() - _started_at
            while status not in [dataset_status.value for dataset_status in dataset_statuses]:
                if _time_since_start > timeout_seconds:
                    raise SuperwiseTimeoutError(
                        f"Timed out while waiting for dataset with ID: '{dataset_id}' to be in one of "
                        f"the statuses {[s.value for s in dataset_statuses]}, "
                        f"got status: '{self._get_field(dataset_id, 'status')}', "
                        f"with reason: '{self._get_field(dataset_id, 'status_reason') or '<unknown reason>'}'"
                    )
                try:
                    dataset = self.get_by_id(dataset_id, timeout=10)
                    status = dataset.status
                except ReadTimeout:
                    self.logger.exception(f"Request for getting dataset:{dataset_id} timed out")
                time.sleep(5)
                _time_since_start = time.perf_counter() - _started_at

            return dataset

        return __wait_to_status

    def _trigger_full_flow(
        self,
        model: Dataset,
        return_model: bool,
        wait_until_complete: bool,
        timeout_seconds: int,
        on_failure: str,
        **kwargs,
    ):
        model.full_flow = True
        response = BaseController.create(self, model, return_model, **kwargs)
        response_id = response.id if return_model else response["id"]

        if not wait_until_complete:
            return response

        wait_to_status = self._get_timeout_func(
            timeout_seconds=timeout_seconds,
            dataset_id=response_id,
            dataset_statuses=[DatasetStatus.SUMMARIZED, DatasetStatus.FAILED],
        )
        try:
            self.logger.info("Waiting for the dataset to be completely processed...")
            dataset = wait_to_status()
        except SuperwiseTimeoutError:
            if on_failure == DatasetOnFailureOptions.RAISE:
                raise
            return response
        if dataset.status == DatasetStatus.FAILED.value and on_failure == DatasetOnFailureOptions.RAISE:
            raise SuperwiseDatasetFailureError(
                f"Dataset creation failed for ID '{response_id}'. "
                f"got status: '{self._get_field(response_id, 'status')}', "
                f"with reason: '{self._get_field(response_id, 'status_reason') or '<unknown reason>'}'"
            )

        return dataset

    def create(
        self,
        model: Dataset,
        return_model=True,
        gcs_service_account: dict = None,
        aws_access_key_id: str = None,
        aws_secret_access_key: str = None,
        aws_role_arn: str = None,
        azure_connection_string: str = None,
        wait_until_complete: bool = True,
        timeout_seconds: int = 60 * 5,
        on_failure="raise",
        **kwargs,
    ):
        """
        ### Description:

        Create a new dataset.

        ### Args:

        `model`: Dataset model.

        `return_model`: return model if True or response.body if False. Default True.

        `gcs_service_account`: GCP service account object used to authenticate and pull dataset files from a customer
                               GCS bucket. If not provided, will be inferred from the environment.
                               (See Google Cloud auth)

        `aws_access_key_id`: AWS access key ID used to authenticate and pull dataset files from a customer S3 bucket.
                             If not provided, will be inferred from the environment.
                             (Used together with `aws_secret_access_key` parameter)

        `aws_secret_access_key`: AWS secret access key used to authenticate and pull dataset files from a customer S3
                                 bucket. If not provided, will be inferred from the environment.
                                 (Used together with `aws_access_key_id` parameter)

        `aws_role_arn`: AWS role ARN used to authenticate and pull dataset files from a customer S3 bucket.
                        If not provided, the authentication will use the `aws_access_key_id` and `aws_secret_access_key`
                        parameters.

        `azure_connection_string`: Azure blob storage connection string used to authenticate and pull dataset files from
                                   a customer blob storage container.
                                   MUST be provided in order to pull files from azure.

        `wait_until_complete`: if True, wait until the dataset is fully processed in the system, and return the final
                               object. If False, return immediately after the dataset is created and the given dataset
                               files are validated, without waiting for the processing. A partially set Dataset object
                               is returned, without all the processed fields. Afterwards the status can be checked with
                               'get_by_id' method. Default True.

        `timeout_seconds`: Timeout for dataset processing waiting. Only relevant if 'wait_until_complete' is True.
                           Default 5 minutes.

        `on_failure`: Action to take in case the dataset processing failed. Only relevant if 'wait_until_complete'
                      is True. Possible values are:
                      - 'ignore': Don't raise an exception, and return the object.
                      - 'raise': Raise a 'SuperwiseDatasetFailureError' exception.
                      Default 'raise'.
        """
        if not isinstance(model.files, list):
            model.files = [model.files]
        on_failure = DatasetOnFailureOptions(on_failure)

        self.logger.info("Uploading dataset files...")
        file_paths = list(
            self._upload_files_to_storage(
                file_paths=model.files,
                project_id=model.project_id,
                gcs_service_account=gcs_service_account,
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                aws_role_arn=aws_role_arn,
                azure_connection_string=azure_connection_string,
            )
        )
        model.internal_files = self._extract_inner_paths(file_paths)
        model.files = self._extract_external_paths(file_paths)

        self.logger.info("Finished uploading, start processing dataset...")
        return self._trigger_full_flow(model, return_model, wait_until_complete, timeout_seconds, on_failure, **kwargs)



import os
import shutil
from pathlib import Path

DOCS_PATH = "temp/"
NOT_ALLOWED_FILES = ["temp/config.md", "temp/utils/managed_env.md", "temp/utils/index.md"]
CURRENT_DIR = os.getcwd()
OUTPUT_DIR = "docs/"


def is_files_not_allowed(file):
    return file in NOT_ALLOWED_FILES


def get_all_files():
    files = []
    for r, d, f in os.walk(DOCS_PATH):
        for file in f:
            if file.endswith(".md"):
                files.append((os.path.join(r, file)))
    return files


def handle_line(line: str):
    new_line = line.strip()
    if new_line.startswith("#"):
        new_line = "#" + new_line
    if new_line.startswith("======================="):
        new_line = new_line.replace("=", "-")
    if new_line.endswith("`\n"):
        new_line = new_line + " \n"
    if new_line.startswith(": ###"):
        new_line = new_line.split(": ###")[1]
    if new_line.startswith(":   ###"):
        new_line = new_line.split(":   ###")[1]
    if new_line.startswith(":"):
        new_line = new_line.replace(":", "")
    new_line += " \n"
    return new_line


def copy_static_files():
    os.rename(f"{CURRENT_DIR}/{DOCS_PATH}index.md", f"{CURRENT_DIR}/{DOCS_PATH}superwise.md")
    shutil.copyfile(f"{CURRENT_DIR}/index.md", f"{CURRENT_DIR}/{OUTPUT_DIR}index.md")
    shutil.copyfile(f"{CURRENT_DIR}/CHANGELOG.md", f"{CURRENT_DIR}/{OUTPUT_DIR}CHANGELOG.md")
    shutil.copytree(f"{CURRENT_DIR}/assets", f"{CURRENT_DIR}/{OUTPUT_DIR}assets")


def write_file(new_file, new_file_url):
    print(f"writing to file {new_file_url}")
    output_file = Path(new_file_url)
    output_file.parent.mkdir(exist_ok=True, parents=True)
    output_file.write_text(new_file)


def main():
    copy_static_files()
    files = get_all_files()
    for file in files:
        if is_files_not_allowed(file):
            continue
        new_file = ""
        data = open(file, "r").readlines()
        for line in data:
            new_line = handle_line(line)
            new_file += new_line
        file_name = ("/").join(file.split("/")[1:])
        new_file_url = f"{CURRENT_DIR}/{OUTPUT_DIR}{file_name}"
        write_file(new_file, new_file_url)


if __name__ == "__main__":
    main()



""" This module contain BaseModel class  """
import datetime
from enum import Enum

import pandas as pd


class BaseModel:
    """Base model"""

    def __init__(self, **kwargs):
        """
        ### Description:

        Constructor of BaseModel class

        """
        self.id = kwargs.get("id")

    def from_datetime(self, date):
        """
         ### Description:

         Convert datetime helper

         ### Args:

         `date`: a datetime.datetime object

        ### Return:

        string of date in iso format
        """

        if pd.isna(date):
            return None
        elif isinstance(date, datetime.datetime):
            return date.isoformat()
        else:
            return date

    def get_properties(self):
        """
        ### Description:

        Get properties of model as a dictionary

        ### Args:

        `date`: a datetime.datetime object

        ### Return:

        dictionary of model properties
        """
        return dict(
            (name, getattr(self, name))
            for name in dir(self)
            if not name.startswith("__") and not callable(getattr(self, name))
        )

    def __repr__(self):
        properties = self.get_properties()
        prop_str = ""
        for p in properties:
            prop_str += f"\t{p}: {properties[p]}\n"
        return f"<Superwise.models.{self.__class__.__name__} object> " f"{{\n" f"{prop_str}" f"}}"



from http import HTTPStatus

import jwt
import pytest
import requests
from _pytest.monkeypatch import MonkeyPatch
from cryptography.hazmat.primitives.asymmetric import rsa
from google.cloud import storage
from google.oauth2 import service_account
from requests import Response

from superwise import Client
from superwise import Superwise
from superwise.utils.storage.internal_storage.aws import AWSInternalStorage
from superwise.utils.storage.internal_storage.azure import AzureInternalStorage
from superwise.utils.storage.internal_storage.gcs import GCSInternalStorage
from superwise.utils.storage.internal_storage.internal_storage import CloudProvider
from superwise.utils.storage.internal_storage.internal_storage import InternalStorage

monkey_patch = MonkeyPatch()


@pytest.fixture(scope="function")
def mock_gcp_internal_bucket(monkeypatch):
    monkeypatch.setattr(
        GCSInternalStorage,
        "_exchange_jwt_for_access_token",
        lambda *args, **kwargs: "access token",
    )

    monkeypatch.setattr(GCSInternalStorage, "_is_cloud_token_active", lambda *args, **kwargs: True)
    monkeypatch.setattr(GCSInternalStorage, "_put_in_storage", lambda *args, **kwargs: "gs://somepath")


@pytest.fixture(scope="function")
def mock_refresh_internal_bucket_token(monkeypatch):
    def mock_refresh_aws(self, *args, **kwargs):
        self._cloud_token = {
            "url": "https://test.s3.amazonaws.com/",
            "fields": {
                "key": "landing/test.csv",
                "AWSAccessKeyId": "asdasdasdasd",
                "x-amz-security-token": "adasdasdasd",
                "signature": "asdasd=",
            },
        }

    def mock_refresh_general(self, *args, **kwargs):
        private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
        self._cloud_token = jwt.encode({"payload": "aaaaa"}, private_key, algorithm="RS256")

    monkeypatch.setattr(InternalStorage, "_refresh_cloud_token", mock_refresh_general)

    monkeypatch.setattr(AWSInternalStorage, "_refresh_cloud_token", mock_refresh_aws)


@pytest.fixture(scope="function")
def mock_azure_internal_bucket(monkeypatch):
    put_response = Response()
    put_response.status_code = HTTPStatus.OK
    monkeypatch.setattr(AzureInternalStorage, "_is_cloud_token_active", lambda *args, **kwargs: True)
    put_response = Response()
    put_response.status_code = 200
    monkeypatch.setattr(requests, "put", lambda *args, **kwargs: put_response)


@pytest.fixture(scope="function")
def mock_get_token(monkeypatch):
    def _mock_token(self):
        return {
            "expires": "Thu, 29 Sep 2022 08:00:41 GMT",
            "expiresIn": 86400,
            "accessToken": "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCIsImtpZCI6IjlkYmRlNzNkIn0.eyJzdWIiOiI0NGE4ODc3YS0yOWJhLTQ2NTYtOGY1OC1lMTQzYTM4MTQ5MTMiLCJlbWFpbCI6InN3X3Rlc3RzQHN1cGVyd2lzZS5haSIsInVzZXJNZXRhZGF0YSI6eyJvbmJvYXJkaW5nIjp0cnVlfSwidGVuYW50SWQiOiJ0ZXN0cyIsInJvbGVzIjpbIlVzZXIiXSwicGVybWlzc2lvbnMiOlsiZmUuc2VjdXJlLndyaXRlLnVzZXJBcGlUb2tlbnMiLCJmZS5zZWN1cmUuZGVsZXRlLnVzZXJBcGlUb2tlbnMiLCJmZS5zZWN1cmUucmVhZC51c2VyQXBpVG9rZW5zIiwiZmUuc2VjdXJlLnJlYWQudGVuYW50QXBpVG9rZW5zIl0sIm1ldGFkYXRhIjp7fSwiY3JlYXRlZEJ5VXNlcklkIjoiZTAyNGE1ZTItNzZkMS00N2YyLWE2NWUtMDViMTFjNzI2MDZlIiwidHlwZSI6InVzZXJBcGlUb2tlbiIsInVzZXJJZCI6ImUwMjRhNWUyLTc2ZDEtNDdmMi1hNjVlLTA1YjExYzcyNjA2ZSIsImF1ZCI6IjlkYmRlNzNkLWViODktNGMxMC05ZDY1LTYwNTI1ZGE0ZDlkNiIsImlzcyI6ImZyb250ZWdnIiwiaWF0IjoxNjY0MzUyMDQxLCJleHAiOjE2NjQ0Mzg0NDF9.UWEqljuU1txF5VDVIqJO8zDY10myPuYqIwx72ClriJh4fAW84jypeDoopVfD01kKHeZDGwJ9YmlGmc6rwQij7YPjXvMRlBSWdshz-2qM7rhY8IOuSBUsW-GgS9BUTKL14K1vbpZ6rLpTEJNoeAoIUKxztynbXhAZVEci78Q_bM29Bs-urSHrzy_hu0hrzAjDKyNkHkNB0DYLiqf03SzyuJDNlj-fvcXp0EaIeqT1FOcpSeJqPgJ57JXrsxg8o_IwZsN-AR_aGdb-zyw0hVwJCzF1wNWMUROo-o2yFBVBwlbh4PMBQkkTA0hiNNqMPI7rA45c2NVicok_e0bi6ZlYkA",
            "refreshToken": "26323edd-8551-422b-99b6-353c76139359",
        }

    monkeypatch.setattr(Client, "get_token", _mock_token)


@pytest.fixture
def mock_gcp_client(monkeypatch):
    parametrized_mock_gcp_client(1000)


def parametrized_mock_gcp_client(size):
    class GCSClient:
        def __init__(self, *args, **kwargs):
            self.name = "test"
            self.size = size

        def bucket(self, bucket_name):
            return GCSClient()

        def get_blob(self, file_name):
            return GCSClient()

        def blob(self, file_name):
            return GCSClient()

        def download_as_string(self):
            return "asdasdaas"

        def upload_from_string(self, data):
            return None

    monkey_patch.setattr(service_account.Credentials, "from_service_account_info", lambda *args, **kwargs: "")
    monkey_patch.setattr(storage, "Client", lambda *args, **kwargs: GCSClient())


# running in all cloud providers environments
@pytest.fixture(scope="function", params=[provider.value for provider in CloudProvider])
def mock_admin_settings(monkeypatch, request):
    settings = {
        "file_upload_settings": {
            "cloud_provider": request.param,
            "azure_account_name": "test",
            "azure_container_name": "test",
            "gcs_bucket_name": "test",
            "s3_bucket_name": "test",
        }
    }
    monkeypatch.setattr(Superwise, "_get_admin_settings", lambda *args, **kwargs: settings)


@pytest.fixture(scope="function")
def sw(
    mock_admin_settings,
    mock_get_token,
    mock_refresh_internal_bucket_token,
    mock_gcp_internal_bucket,
    mock_azure_internal_bucket,
):
    return Superwise(client_id="test", secret="test")



""" This module implement gcs client functionality  """
import google.api_core.exceptions
from google.cloud import storage
from google.oauth2 import service_account as gcp_service_account

from superwise import logger
from superwise.utils.exceptions import SuperwiseFileTooLargeError
from superwise.utils.exceptions import SuperwiseStorageDownloadGCSError
from superwise.utils.exceptions import SuperwiseStorageUploadGCSError
from superwise.utils.singelton import singleton
from superwise.utils.storage import extract_directory
from superwise.utils.storage import validate_file_size


@singleton
class GcsClient:
    def __init__(self):
        self.logger = logger

    def get_file_from_gcs(self, file_path, service_account: dict = None, validate_size: bool = False) -> dict:
        """
        ### Description:

        get the content of file from gcs

        ### Args:

        `file_path`:  the file patch

        `service_account`:  ssuperwise service account object

        ### Return:

        a string of the object (file) content
        """
        if not str(file_path).startswith("gs://"):
            self.logger.error("Failed upload file to superwise storage")
            raise Exception("file_path must start with 'gs://'")
        self.logger.info("Download file {} from gcs".format(file_path))
        bucket, key = extract_directory(file_path)
        try:
            customer_bucket = self.create_gcs_bucket_connection(bucket, service_account)
            blob = customer_bucket.get_blob(key)
            size = self._validate_blob_size(file_path, blob) if validate_size else None
            data = blob.download_as_string()
        except SuperwiseFileTooLargeError:
            raise
        except Exception as e:
            self.logger.error(f"Error download file {file_path} from gcs with ext{e}")
            raise SuperwiseStorageDownloadGCSError(f"Error download file {file_path} from gcs")
        return dict(filename=key, data=data, size=size)

    def create_gcs_bucket_connection(self, bucket_name: str, service_account: dict = None):
        """
        ### Description:

        get connection to gcs bucket

        ### Args:

        `bucket_name`: gcs bucket name

        `service_account`: superwise service account object

        """
        try:
            self.logger.debug(f"Create connection to superwise bucket {bucket_name}")
            if service_account:
                credentials = gcp_service_account.Credentials.from_service_account_info(service_account)
                gcs_client = storage.Client(credentials=credentials)
            else:
                gcs_client = storage.Client()
            bucket_connection = gcs_client.bucket(bucket_name)
            return bucket_connection
        except Exception as e:
            self.logger.error(f"Error create connection to superwise bucket {bucket_name}")
            raise Exception(f"Error create connection to superwise bucket {bucket_name}")

    def upload_string_to_bucket(self, data, blob_name, bucket):
        try:
            self.logger.debug(f"Upload file to superwise bucket {blob_name}")
            blob = bucket.blob(blob_name)
            blob.upload_from_string(data=data)
            return f"gs://{bucket.name}/{blob_name}"
        except google.api_core.exceptions.Forbidden as e:
            raise SuperwiseStorageUploadGCSError(f"Failed upload file to superwise storage {blob_name} with ext {e}")
        except Exception as e:
            self.logger.error(f"Failed upload file to superwise storage {blob_name} with ext {e}")
            raise SuperwiseStorageUploadGCSError(f"Failed upload file to superwise storage {blob_name} with ext {e}")

    @staticmethod
    def _validate_blob_size(filename, blob):
        logger.debug(f"Blob {filename} size is {blob.size}")
        size = blob.size

        validate_file_size(file_path=filename, file_size=size)
        return size



import boto3
import pytest
import requests
from google.cloud import storage
from google.oauth2 import service_account
from requests import Response

from superwise import Superwise
from superwise.models.segment import Segment
from superwise.models.segment import SegmentCondition
from superwise.models.segment import SegmentConditionDefinition
from superwise.utils.client import Client
from superwise.utils.exceptions import *
from tests import get_sw
from tests import print_results

segment_id = None


def mock_boto_client(monkeypatch):
    class BotoClient:
        def __init__(self, *args, **kwargs):
            pass

        def get_object(self, *args, **kwargs):
            class ObjectResposne:
                def read(self):
                    return "blablabla"

            return dict(Body=ObjectResposne())

    monkeypatch.setattr(boto3, "client", lambda *args, **kwargs: BotoClient())


@pytest.fixture(scope="function")
def mock_gcp_client(monkeypatch):
    class GCSClient:
        def __init__(self, *args, **kwargs):
            self.name = "test"

        def bucket(self, bucket_name):
            return GCSClient()

        def blob(self, file_name):
            return GCSClient()

        def download_as_string(self):
            return "asdasdaas"

        def upload_from_string(self, data):
            return None

    monkeypatch.setattr(service_account.Credentials, "from_service_account_info", lambda *args, **kwargs: "")
    monkeypatch.setattr(storage, "Client", lambda *args, **kwargs: GCSClient())


@pytest.fixture(scope="function")
def mock_get_token(monkeypatch):
    def _mock_token(self):
        return {
            "expires": "Thu, 29 Sep 2022 08:00:41 GMT",
            "expiresIn": 86400,
            "accessToken": "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCIsImtpZCI6IjQ3ZDdmMDg2In0.eyJzdWIiOiI5YzNlZmUxZC03NGNlLTRlZTItYTMyOC1kMWZmNmQyMDAyM2YiLCJlbWFpbCI6InN3X2JhcmFrQHN1cGVyd2lzZS5haSIsInVzZXJNZXRhZGF0YSI6e30sInRlbmFudElkIjoiYmFyYWsiLCJyb2xlcyI6WyJWaWV3ZXIiXSwicGVybWlzc2lvbnMiOlsiZmUuc2VjdXJlLndyaXRlLnVzZXJBcGlUb2tlbnMiLCJmZS5zZWN1cmUuZGVsZXRlLnVzZXJBcGlUb2tlbnMiLCJmZS5zZWN1cmUucmVhZC51c2VyQXBpVG9rZW5zIl0sIm1ldGFkYXRhIjp7fSwiY3JlYXRlZEJ5VXNlcklkIjoiNDg5ZmM5Y2YtZDlhYy00MWMwLWJmM2ItN2VhNDUyNDY4ODEyIiwidHlwZSI6InVzZXJBcGlUb2tlbiIsInVzZXJJZCI6IjQ4OWZjOWNmLWQ5YWMtNDFjMC1iZjNiLTdlYTQ1MjQ2ODgxMiIsImlhdCI6MTYzNjY0ODIyMywiZXhwIjoxNjM2NzM0NjIzLCJpc3MiOiJmcm9udGVnZyJ9.qhEclIsSpfwXpCTFb8qhKpizRWtpQSnkE7VMsy9Et3guLcOcTiTVZ2wOJPmemtL3g3AStKH2jFSOEwQOoqnvgSR3dum9I_Ae3UwrFNRnM3EqOz7UsD0cJAd1AYy-69-67o5oX9A2U4MPZSA5Dr5Edbvn86-AsBJhADGDs5AyEyuGmlJTq0ACGAmoC8qZlxwnOsn9wIzTiQVU7085M73n5iJ26SNhsy4KNpU-8oR2lC1akDroHzL8aIr5dAWSWZz_cfcyWQyC1gqb4_ZAvG1GXiKwsGW2irFyfGoD9zrwMoMGuWXKCbXnHxIzuv8ImX_cRVPXq5xVBYUXwODr83Q3FA",
            "refreshToken": "26323edd-8551-422b-99b6-353c76139359",
        }

    monkeypatch.setattr(Client, "get_token", _mock_token)


@pytest.fixture(scope="function")
def mock_segments_requests(monkeypatch):
    post_response = Response()
    post_response._content = b'{"status": "ACTIVE", "created_at": "2022-08-01 11:39:17", "project_id": 1, "created_by": "auto", "archived_at": null, "id": 7, "definition_json": [[{"value": 23.0, "condition": "==", "entity_id": 4, "entity_name": "age"}]], "name": "hello2"}'
    post_response.status_code = 201
    get_response = Response()
    get_response._content = b'[{"status": "ACTIVE", "created_at": "2022-08-01 11:39:17", "project_id": 1, "created_by": "auto", "archived_at": null, "id": 7, "definition_json": [[{"value": 23.0, "condition": "==", "entity_id": 4, "entity_name": "age"}]], "name": "hello2"}, {"status": "ACTIVE", "created_at": "2022-08-01 11:39:17", "project_id": 1, "created_by": "auto", "archived_at": null, "id": 7, "definition_json": [[{"value": 23.0, "condition": "==", "entity_id": 4, "entity_name": "age"}]], "name": "hello2"}]'
    get_response.status_code = 200
    monkeypatch.setattr(requests, "post", lambda *args, **kwargs: post_response)
    monkeypatch.setattr(requests, "get", lambda *args, **kwargs: get_response)


def test_create_segment_inline(mock_get_token, mock_segments_requests, sw):
    inline_model_test = sw.segment.create(
        Segment(
            project_id=1,
            name="hello2",
            definition=[
                [
                    SegmentConditionDefinition(
                        entity_id=1, condition=SegmentCondition.IN, value=["Israel", "United States of America"]
                    ),
                    SegmentConditionDefinition(entity_id=2, condition=SegmentCondition.GREATER_THAN, value=0.5),
                ]
            ],
        )
    )
    print_results("created segment object 1", inline_model_test.get_properties())
    assert inline_model_test.project_id == 1
    assert inline_model_test.name == "hello2"
    assert inline_model_test.id is not None


def test_get_segment(mock_get_token, mock_segments_requests, sw):
    segments = sw.segment.get({})
    assert len(segments) == 2
    assert int(segments[0].id) == 7



""" This module implement roles functionality  """
from superwise.controller.base import BaseController
from superwise.models.model import Model


class RoleController(BaseController):
    """Role controller class, implement functionalities for roles API"""

    def __init__(self, client, sw):
        """

        ### Args:

        `client`: superwise client object

        `sw`: superwise object

        """

        super().__init__(client, sw)
        self.path = "model/v1/roles"
        self.model_name = "Role"

    def list_to_dict(self, model_list):
        """
        ### Description:

        convert a list of models into dict

        ### Args:

        `model_list`: list of models

        ### Return:

         Dict of RoleModel()
        """

        dict = {}
        for model in model_list:
            properties = model.get_properties()
            dict[properties["description"]] = properties
        return dict



""" This module implement Model model  """
import warnings

from superwise.models.base import BaseModel


class Model(BaseModel):
    """Model model class"""

    def __init__(
        self,
        id=None,
        project_id=None,
        external_id=None,
        name=None,
        description=None,
        monitor_delay=None,
        time_units=None,
        is_archived=None,
        active_version_id=None,
        **kwargs
    ):
        """
        ### Description:

        Constructor for Model class

        ### Args:

        `id`: id of model

        `project_id`: id of project

        `external_id`: external/secondary identifier, use it to map between your id and superwise

        `name`: name of model

        `description`: description for the model

        `time_units`:

        `is_archive`:
        """
        self.external_id = external_id
        self.project_id = project_id
        self.name = name
        self.id = id or None
        self.description = description
        if monitor_delay is not None:
            warnings.warn("Passing monitor_delay is deprecated, it should be configured for a specific Policy")
        self.time_units = time_units
        self.is_archived = is_archived
        self.active_version_id = active_version_id



import json

import responses

from superwise import Client
from superwise import Config


@responses.activate
def test_token_refresh():
    def request_callback_generator():
        # payload = json.loads(request.body)
        return_403 = True

        def _callback(request):
            nonlocal return_403
            valid_resp_body = {
                "expires": "Thu, 29 Sep 2022 08:00:41 GMT",
                "expiresIn": 86400,
                "accessToken": "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCIsImtpZCI6IjlkYmRlNzNkIn0.eyJzdWIiOiI0NGE4ODc3YS0yOWJhLTQ2NTYtOGY1OC1lMTQzYTM4MTQ5MTMiLCJlbWFpbCI6InN3X3Rlc3RzQHN1cGVyd2lzZS5haSIsInVzZXJNZXRhZGF0YSI6eyJvbmJvYXJkaW5nIjp0cnVlfSwidGVuYW50SWQiOiJ0ZXN0cyIsInJvbGVzIjpbIlVzZXIiXSwicGVybWlzc2lvbnMiOlsiZmUuc2VjdXJlLndyaXRlLnVzZXJBcGlUb2tlbnMiLCJmZS5zZWN1cmUuZGVsZXRlLnVzZXJBcGlUb2tlbnMiLCJmZS5zZWN1cmUucmVhZC51c2VyQXBpVG9rZW5zIiwiZmUuc2VjdXJlLnJlYWQudGVuYW50QXBpVG9rZW5zIl0sIm1ldGFkYXRhIjp7fSwiY3JlYXRlZEJ5VXNlcklkIjoiZTAyNGE1ZTItNzZkMS00N2YyLWE2NWUtMDViMTFjNzI2MDZlIiwidHlwZSI6InVzZXJBcGlUb2tlbiIsInVzZXJJZCI6ImUwMjRhNWUyLTc2ZDEtNDdmMi1hNjVlLTA1YjExYzcyNjA2ZSIsImF1ZCI6IjlkYmRlNzNkLWViODktNGMxMC05ZDY1LTYwNTI1ZGE0ZDlkNiIsImlzcyI6ImZyb250ZWdnIiwiaWF0IjoxNjY0MzUyMDQxLCJleHAiOjE2NjQ0Mzg0NDF9.UWEqljuU1txF5VDVIqJO8zDY10myPuYqIwx72ClriJh4fAW84jypeDoopVfD01kKHeZDGwJ9YmlGmc6rwQij7YPjXvMRlBSWdshz-2qM7rhY8IOuSBUsW-GgS9BUTKL14K1vbpZ6rLpTEJNoeAoIUKxztynbXhAZVEci78Q_bM29Bs-urSHrzy_hu0hrzAjDKyNkHkNB0DYLiqf03SzyuJDNlj-fvcXp0EaIeqT1FOcpSeJqPgJ57JXrsxg8o_IwZsN-AR_aGdb-zyw0hVwJCzF1wNWMUROo-o2yFBVBwlbh4PMBQkkTA0hiNNqMPI7rA45c2NVicok_e0bi6ZlYkA",
                "refreshToken": "26323edd-8551-422b-99b6-353c76139359",
            }
            headers = {}
            if return_403:
                return_403 = False
                return (403, headers, json.dumps({}))
            return (200, headers, json.dumps(valid_resp_body))

        return _callback

    responses.post(
        "https://auth.superwise.ai/identity/resources/auth/v1/api-token",
        status=200,
        json={
            "expires": "Thu, 29 Sep 2022 08:00:41 GMT",
            "expiresIn": 86400,
            "accessToken": "eyJhbGciOiJSUzI1NiIsInR5cCI6IkpXVCIsImtpZCI6IjlkYmRlNzNkIn0.eyJzdWIiOiI0NGE4ODc3YS0yOWJhLTQ2NTYtOGY1OC1lMTQzYTM4MTQ5MTMiLCJlbWFpbCI6InN3X3Rlc3RzQHN1cGVyd2lzZS5haSIsInVzZXJNZXRhZGF0YSI6eyJvbmJvYXJkaW5nIjp0cnVlfSwidGVuYW50SWQiOiJ0ZXN0cyIsInJvbGVzIjpbIlVzZXIiXSwicGVybWlzc2lvbnMiOlsiZmUuc2VjdXJlLndyaXRlLnVzZXJBcGlUb2tlbnMiLCJmZS5zZWN1cmUuZGVsZXRlLnVzZXJBcGlUb2tlbnMiLCJmZS5zZWN1cmUucmVhZC51c2VyQXBpVG9rZW5zIiwiZmUuc2VjdXJlLnJlYWQudGVuYW50QXBpVG9rZW5zIl0sIm1ldGFkYXRhIjp7fSwiY3JlYXRlZEJ5VXNlcklkIjoiZTAyNGE1ZTItNzZkMS00N2YyLWE2NWUtMDViMTFjNzI2MDZlIiwidHlwZSI6InVzZXJBcGlUb2tlbiIsInVzZXJJZCI6ImUwMjRhNWUyLTc2ZDEtNDdmMi1hNjVlLTA1YjExYzcyNjA2ZSIsImF1ZCI6IjlkYmRlNzNkLWViODktNGMxMC05ZDY1LTYwNTI1ZGE0ZDlkNiIsImlzcyI6ImZyb250ZWdnIiwiaWF0IjoxNjY0MzUyMDQxLCJleHAiOjE2NjQ0Mzg0NDF9.UWEqljuU1txF5VDVIqJO8zDY10myPuYqIwx72ClriJh4fAW84jypeDoopVfD01kKHeZDGwJ9YmlGmc6rwQij7YPjXvMRlBSWdshz-2qM7rhY8IOuSBUsW-GgS9BUTKL14K1vbpZ6rLpTEJNoeAoIUKxztynbXhAZVEci78Q_bM29Bs-urSHrzy_hu0hrzAjDKyNkHkNB0DYLiqf03SzyuJDNlj-fvcXp0EaIeqT1FOcpSeJqPgJ57JXrsxg8o_IwZsN-AR_aGdb-zyw0hVwJCzF1wNWMUROo-o2yFBVBwlbh4PMBQkkTA0hiNNqMPI7rA45c2NVicok_e0bi6ZlYkA",
            "refreshToken": "26323edd-8551-422b-99b6-353c76139359",
        },
    )
    responses.add_callback(
        responses.POST,
        "https://https//something.com/tests//something",
        callback=request_callback_generator(),
        content_type="application/json",
    )
    # with responses.RequestsMock() as rsps:
    client = Client(None, "secret", "https://something.com")
    client.post(client.build_url("/something"), params={})



""" This module implement Segment model  """
from datetime import datetime
from typing import List
from typing import Optional
from typing import Union

from superwise.models.base import BaseModel
from superwise.resources.superwise_enums import SegmentCondition
from superwise.resources.superwise_enums import SegmentStatus
from superwise.utils.exceptions import SuperwiseValidationException


class SegmentConditionDefinition:
    def __init__(
        self,
        entity_id: int,
        condition: SegmentCondition,
        value: Union[List[str], List[float], List[bool], float, bool, str],
    ):
        self.entity_id = entity_id
        self.condition = condition
        self.value = value
        assert isinstance(condition, SegmentCondition)

    def to_dict(self):
        return dict(entity_id=self.entity_id, condition=self.condition.value, value=self.value)


class Segment(BaseModel):
    """Segment model class"""

    def __init__(
        self,
        id: int = None,
        project_id: int = None,
        name: str = None,
        status: SegmentStatus = None,
        definition: Optional[List[List[SegmentConditionDefinition]]] = None,
        definition_json: str = None,
        created_at: datetime = None,
        created_by: str = None,
        archived_at: datetime = None,
        **kwargs,
    ):
        """
        ### Description:

        Constructor of Segment model class

        ### Args:

        `id`: id of segment

        `name`: name of segment (string)

        `status`: status of the model

        `definition`: definition of the query: list of list SegmentConditionDefinition or the condition in the
         inner list will have OR operator between them and all the inner list will have AND operator between them

        `definition_query`:

        `created_at`: the creation time of the segment

        `created_by`: the user created the segment

        `archived_at`: the archive time of the segment if exists

        """

        self.id = id
        self.project_id = project_id
        self.name = name
        self.status = status

        if definition is not None:
            self.definition_json = []
            for or_block_of_conditions in definition:
                if not isinstance(or_block_of_conditions, list):
                    raise SuperwiseValidationException(
                        "definition should be have be type of List[List[SegmentConditionDefinition]]"
                    )
                or_block_of_dict_conditions = []
                for condition in or_block_of_conditions:
                    if isinstance(condition, SegmentConditionDefinition):
                        or_block_of_dict_conditions.append(condition.to_dict())
                    else:
                        raise SuperwiseValidationException(f"{condition} is not instance of SegmentConditionDefinition")
                self.definition_json.append(or_block_of_dict_conditions)
        else:
            self.definition_json = definition_json
        self.created_at = created_at
        self.created_by = created_by
        self.archived_at = archived_at



""" This module implement Transaction model  """
from typing import List
from typing import Optional
from typing import Union

from superwise.models.base import BaseModel


class Transaction(BaseModel):
    """Transaction model"""

    def __init__(
        self,
        id: Optional[int] = None,
        model_id: Optional[int] = None,
        model_name: Optional[str] = None,
        version_id: Optional[Union[str, int]] = None,
        origin_url: Optional[str] = None,
        transaction_id: str = None,
        integration_type: Optional[str] = None,
        file_type: Optional[str] = None,
        status: Optional[str] = None,
        created_at: int = None,
        details: Optional[str] = None,
        num_of_records: Optional[str] = None,
        is_reviewed: bool = None,
        metadata: dict = None,
        **kwargs
    ):
        """
        ### Description:

        Constructor for Transaction class

        ### Args:

        `id`: id of model

        `model_id`: id of model

        `model_name`: name of the model

        `version_id`: id of the version

        `origin_url`: url of source file (Optional[str])

        `transaction_id`: transaction id (str)

        `integration_type`: Optional[str]

        `file_type`: Optional[str]

        `status`:  Optional[str]

        `created_at`: timestamp (int)

        `details`: Optional[str]

        `num_of_records`: Optional[str]

        `is_reviewed`: bool
        """
        self.id = id
        self.model_id = model_id
        self.model_name = model_name
        self.version_id = version_id
        self.origin_url = origin_url
        self.transaction_id = transaction_id
        self.integration_type = integration_type
        self.file_type = file_type
        self.status = status
        self.created_at = created_at
        self.details = details
        self.num_of_records = num_of_records
        self.is_reviewed = is_reviewed
        self.metadata = metadata



""" This module implement version functionality  """
import time

from superwise.config import Config
from superwise.controller.base import BaseController
from superwise.models.data_entity import DataEntity
from superwise.utils.exceptions import SuperwiseValidationException


class VersionController(BaseController):
    """Version controller"""

    def __init__(self, client, sw):
        """
        ### Description:

        Constructor of VersionController

        ### Args:

        `client`:  superwise client object

        `sw`:  superwise object

        """
        super().__init__(client, sw)
        self.path = "model/v1/versions"
        self.model_name = "Version"

    def create(self, model, is_return_model=True, **kwargs):
        """
         ### Description:

         create version

         ### Args:

         `model`:  Version model object

         `is_return_model`:

        `wait_until_complete`:  if set to True, function will blocked untill creation of version is completed

        """

        if len(model.data_entities) and isinstance(model.data_entities[0], DataEntity):
            model.data_entities = [m.get_properties() for m in model.data_entities]
        res = super().create(model, **kwargs)
        return res

    def _dict(self, params, model_name=None):
        model_name = model_name or self.model_name

        model = super()._dict_to_model(params, model_name=model_name)

        if model_name != "DataEntity" and len(model.data_entities) and isinstance(model.data_entities[0], DataEntity):
            model.data_entities = [m.get_properties() for m in model.data_entities]
        return model

    def get_data_entities(self, version_id):

        """
        ### Description:

        Get dataentities of a given version

        ### Args:

        `version_id`:  version id (int)

        ### Return:

        List[DataEntity] - list of DataEntity objects
        """

        models = []
        url = self.client.build_url("{}/{}/data_entities".format(self.path, version_id))
        self.logger.info("CALL GET DATA ENTITIES {} ".format(url))
        r = self.client.get(url)
        entities_lst = self.parse_response(r, "DataEntity", is_return_model=False)
        entities_lst = [e["data_entity"] for e in entities_lst]
        for entity in entities_lst:
            models.append(self._dict_to_model(entity, "DataEntity"))
        return models

    def activate(self, version_id):
        """
        ### Description:

        activate pending version

        ### Args:

        `version_id`:   version id (int)

        ### Return:

        Version object from server
        """

        url = "{}/{}".format(self.path, version_id)
        res = self.patch(url, {"status": "Active"})
        return res



""" This module implement Role model  """
from superwise.models.base import BaseModel


class Role(BaseModel):
    """Role model class, model  for roles data"""

    def __init__(
        self,
        id=None,
        model_type_id=None,
        is_nullable=None,
        description=None,
        internal_name=None,
        is_label=None,
        is_optional=None,
        role=None,
        value=None,
        **kwargs
    ):
        """
        ### Description:

        Constructor of Role class

        ### Args:

        `id`: id of role (int)

        `model_type_id`: model type id (ModelType enum value)

        `is_nullable`:

        `description`: description of the role

         internal_name`:

        `is_label`:

        `is_optional`:

        `role`:

        `value`:

        """
        self.model_type_id = model_type_id
        self.description = description
        self.internal_name = internal_name
        self.is_label = is_label
        self.is_optional = is_optional
        self.value = value
        self.is_nullable = is_nullable
        self.role = role
        self.id = id



import os

from superwise.utils.storage import validate_file_size


def load_from_local_storage(file_path: str, validate_size: bool = False) -> dict:
    file_size = None
    if validate_size:
        file_size = os.path.getsize(file_path)
        validate_file_size(file_path=file_path, file_size=file_size)

    with open(file_path, "rb") as file:
        data = file.read()

    filename = file_path.split("/")[-1]

    return dict(filename=filename, data=data, size=file_size)



""" This module implement tasks functionality  """
import json
from typing import Dict
from typing import List

import pandas as pd
from pandas import DataFrame
from requests import Response

from superwise.controller.base import BaseController
from superwise.models.notification import Notification
from superwise.resources.superwise_enums import NotifyUpon
from superwise.resources.superwise_enums import ScheduleCron


class PolicyController(BaseController):
    """Class ModelController - responsible for task functionality"""

    def __init__(self, client, sw):
        """

        ### Args:

        `client`: superwise client object

        `sw`: superwise  object

        """
        super().__init__(client, sw)
        self.path = "monitor/v1/policy"
        self.model_name = "Policy"

    def get_policy_templates(self) -> DataFrame:
        """
        ### Description:
        Get all the available templates
        """

        response = self.client.get(self.client.build_url(f"{self.path}/templates"))

        assert response.text != "" and response.text != "", "Something went wrong: 0 templates received"

        fields_to_display = ["name", "description"]
        templates: List[Dict] = response.json()

        return pd.DataFrame(templates)[fields_to_display]

    def create_policy_from_template(
        self,
        model_id: int,
        template_name: str,
        notify_upon: NotifyUpon,
        notification_channels: List[Notification],
        schedule: ScheduleCron,
        policy_name: str = None,
        minimal_quantity: int = 0,
        monitor_offset: int = 1,
    ) -> Response:
        """
        ### Description:
        Used for creation of policy from one of the available templates

        ### Args:

        `model_id`: id of the model which the policy will monitor

        `template_name`: The name of the template which the policy will be created from

        `notify_upon`: When you should be notified

        `notification_channels`: List of all the notifications channels you would like to receive notification when the
         policy will find violations

        `schedule`: Cronstring - defines when the policy will be run and data will be sampled for violations.

        `policy_name`: Optional - The name of the policy that will be created, if not supplied the name will be the
        taken from the template name and the model's id

        `minimal_quantity`: Optional - Minimum threshold value for ignoring sampling/aggregation of time frames with
         less than
        the threshold of rows.

        `monitor_offset`: Optional - Amount of days for delay until they are monitored, useful when you need to add
         label data or
        wait until all the data is available.

        """
        notification_channels_ids = list(map(lambda notification: notification.id, notification_channels))
        payload = {
            "template_name": template_name,
            "notify_upon": notify_upon.value,
            "notification_channels": notification_channels_ids,
            "model_id": model_id,
            "schedule": schedule.value,
            "policy_name": policy_name,
            "minimal_quantity": minimal_quantity,
            "monitor_offset": monitor_offset,
        }

        return self.client.post(self.client.build_url(f"{self.path}/templates"), payload)



from typing import Union

import humanize

from superwise.config import Config
from superwise.utils.exceptions import SuperwiseFileTooLargeError


def extract_directory(url: str):
    """
    ### Description:

    extract  given url of gcs to bucket name and prefix

    ### Args:

    `url`:  the url to extract

    ### Return:

    a tuple of bucket and prefix
    """
    bucket = url.split("/")[2]
    prefix = "/".join(url.split("/")[3:])
    return bucket, prefix


def validate_file_size(file_path, file_size):
    if file_size > Config.FILE_SIZE_LIMIT_BYTES + Config.FILE_SIZE_PADDING:
        readable_file_size = humanize.filesize.naturalsize(file_size)
        readable_maximum_size = humanize.filesize.naturalsize(Config.FILE_SIZE_LIMIT_BYTES)
        raise SuperwiseFileTooLargeError(
            f"File '{file_path}' is too large: ({readable_file_size}), maximum size allowed is {readable_maximum_size}"
        )


def validate_files_size(filenames: Union[list, str], size: int):
    if not isinstance(filenames, list):
        filenames = [filenames]

    if size > Config.FILE_SIZE_LIMIT_BYTES + Config.FILE_SIZE_PADDING:
        readable_size = humanize.filesize.naturalsize(size)
        readable_maximum_size = humanize.filesize.naturalsize(Config.FILE_SIZE_LIMIT_BYTES)
        raise SuperwiseFileTooLargeError(
            f"The *total* size of the files: {filenames} is too large: ({readable_size}), "
            f"maximum size allowed is {readable_maximum_size}"
        )



""" This module implements data entities functionality """
import pandas as pd

from superwise.config import Config
from superwise.controller.base import BaseController
from superwise.controller.infer import infer_dtype
from superwise.controller.summary.entities_validator import EntitiesValidator
from superwise.controller.summary.feature_importance import FeatureImportance
from superwise.controller.summary.summary import Summary
from superwise.models.data_entity import DataEntity
from superwise.models.data_entity import DataEntitySummary
from superwise.resources.superwise_enums import get_enum_value
from superwise.utils.exceptions import SuperwiseUnsupportedException
from superwise.utils.exceptions import SuperwiseValidationException


class DataEntityController(BaseController):
    """controller for Data entities"""

    def __init__(self, client, sw):
        """


        ### Args:

        `client`: client object

        `sw`: superwise object

        """
        super().__init__(client, sw)
        self.path = "model/v1/data_entities"
        self.model_name = "DataEntity"
        self._entities_df = None
        self.data = None

    @staticmethod
    def _pre_process_data(data):
        data.columns = data.columns.str.lower()
        for column in Config.LIST_DROP_DATA_COLS:
            if column in data.columns:
                data = data.drop(column, axis=1)
        return data

    def create(self, name=None, type=None, dimension_start_ts=None, role=None, feature_importance=None):
        """
        ### Description:

         create dataentity

        ### Args:

        `name`: name for dataentity

        `type`:

        `dimension_start_ts`:

        `role`: role of dataentity (DataEntityRole)

        `feature_importance`: feature importance value

        """

        params = locals()
        return self._dict_to_model(params)

    def update_summary(self, data_entity_id, summary):
        """
        ### Description:

        update summary implementation
        """
        self.model = DataEntitySummary(data_entity_id, summary)
        self.model_name = "DataEntitySummary"
        self.create(self.model)

    def generate_summary(self, data_entities, model, data, base_version=None, **kwargs):
        """
        ### Description:
        Unsupported anymore, this function raise Exception
        """
        raise SuperwiseUnsupportedException(
            "Unsupported SDK API, please use  summarise() instead," " See CHANGELOG for more details"
        )

    def summarise(
        self,
        data,
        specific_roles,
        default_role="feature",
        entities_dtypes=None,
        importance_mapping=None,
        importance_target_label=None,
        importance_sample=None,
        base_version=None,
        **kwargs
    ):
        """
        ### Description:
        Summarise the baseline data

        ### Args:

        `data`: a dataframe of baseline data

        `entities_dtypes`: ndtypes dictionary or None for auto infer internally

            HOW TO USE OPTIONS:

            - Manually create dictionary as:
                  {"feature_0" : "Boolean",
                "feature_1" : "Categorical"
                }

            -  run infer to get dictionary and pass the value to allow override of infer results:
                from superwise.controller.infer import infer_dtype
                dtypes = infer_dtypes(baseline_df)

            - pass None
                the summarise function will auto generate the dtypes.

            - Note: categorical entities with over 200 categories (e.g first name, street name, IP, etc.) will be counted as "Sparse" features, and will not have any metric calculated on it other than "Missing values"

        `specific_roles`: dictionary of roles, for example:

            {"feature_1" : "feature",
             "record_id" : "id"}

        `default_role`: a default role used by SDK (normally and by default is feature)

        `importance_mapping`: mapping dictionary for feature importance, example:

            {"feature_0" : 0.4}


        `importance_target_label`: option to specifics a label target for feature importance

        `importance_sample`: allow sampling of data for feature importance.
                this option should be used for big baseline data

        `base_version`:  base version. useful to inherit information from already exist version

        ### Return:

        list of DataEntity objects.
        """
        if not entities_dtypes:
            entities_dtypes = infer_dtype(data)
        entities_dtypes = dict((k.lower(), v) for k, v in entities_dtypes.items())
        for key in specific_roles:
            specific_roles[key] = get_enum_value(specific_roles[key])

        self.data = self._pre_process_data(data)
        data_entities = []
        for entity in entities_dtypes:
            importance_value = importance_mapping.get(entity, None) if importance_mapping else None
            data_entities.append(
                {
                    "name": entity,
                    "role": specific_roles.get(entity, default_role),
                    "type": entities_dtypes[entity],
                    "feature_importance": importance_value,
                    "dimension_start_ts": None,
                    "id": None,
                }
            )

        if base_version:
            if base_version.status not in ["Pending", "Active"]:
                raise SuperwiseValidationException(
                    "base version should be used for summarized only version, current status: {}".format(
                        base_version.status
                    )
                )
            r = self.client.get(self.client.build_url("model/v1/versions/{}/data_entities".format(base_version.id)))
            previous_entities = self.parse_response(r, is_return_model=False)
            previous_entities = [e["data_entity"] for e in previous_entities]
            previous_entities_dict = {}
            if previous_entities:
                for data_entity in previous_entities:
                    previous_entities_dict[data_entity["name"]] = data_entity["id"]
            for de in data_entities:
                if de["name"] in previous_entities_dict:
                    de["id"] = previous_entities_dict[de["name"]]
                else:
                    de["id"] = None

        entities_df = pd.DataFrame(data_entities)
        self._entities_df = entities_df
        validator = EntitiesValidator(entities_df, self.data)

        self.data = validator.prepare()
        if not importance_mapping:
            fi = FeatureImportance(self._entities_df)
            self._entities_df = fi.compute(self.data, target=importance_target_label, sample=importance_sample)
        entities_df_summary = Summary(self._entities_df, self.data).generate()
        data_entities = DataEntity.df_to_list(entities_df_summary)
        return data_entities



""" This module implement Notification model  """
import datetime
import warnings

from superwise.models.base import BaseModel


class Notification(BaseModel):
    """Notification model class"""

    def __init__(
        self,
        id: int = None,
        name: str = None,
        type: str = None,
        notification_metadata=None,
        created_at: datetime = None,
        last_delivery_status=None,
        last_delivery_ts: datetime = None,
        is_enabled: bool = True,
        is_hide: bool = False,
        **kwargs
    ):
        """
        ### Description:

        Constructor for Notification class

        ### Args:

        `id`: id of model

        `type`: type of the notification

        `name`: name of notification

        `last_delivery_ts`: description for the model

        `notification_metadata`:

        `created_at`:
        """
        self.id = id
        self.name = name
        self.last_delivery_ts = last_delivery_ts
        self.is_hide = is_hide
        self.is_enabled = is_enabled
        self.type = type
        self.notification_metadata = notification_metadata
        self.created_at = created_at
        self.last_delivery_status = last_delivery_status



import json
import os.path

import boto3
import pytest
import requests
from requests import Response

from project_root import PROJECT_ROOT
from superwise.models.transaction import Transaction
from superwise.utils.exceptions import SuperwiseValidationException
from superwise.utils.storage.azure import AzureClientReader


@pytest.fixture(scope="function")
def mock_transaction_requests(monkeypatch):
    the_response = Response()
    the_response._content = b'{ "transaction_id" : "123" ,"metadata" : {}}'
    the_response.status_code = 201
    get_response = Response()
    get_response._content = b"{\"created_at\": 1639668801319, \"details\": \"predictions columns diff: {'email_11174', 'linking_identity_230', 'linking_strong_187', 'linking_identity_637', 'ip_country_code', 'linking_identity_765', 'linking_identity_636', 'is_digital', 'linking_online_755', 'linking_strong_169', 'ip_organization', 'linking_online_519', 'order_total_spent', 'shipping_province_code', 'bin_country_code', 'prediction_probability', 'periodic_310', 'order_83', 'customer_49', 'vendor_id', 'periodic_360', 'currency_code', 'linking_online_744', 'product_v3', 'merchant_id', 'decision_reason', 'technical_plan', 'cc_69', 'ts', 'periodic_135', 'linking_strong_780', 'beacon_309', 'linking_identity_778', 'periodic_8', 'linking_online_520', 'approve_threshold', 'cc_822', 'action_review', 'order_source', 'billing_province_code', 'order_submission_reason', 'cc_823', 'order_language', 'linking_identity_849', 'customer_410', 'cc_223', 'email_186', 'shipping_country_code', 'linking_strong_38', 'order_type', 'linking_identity_31', 'customer_v2', 'order_245', 'ip_region', 'linking_strong_193', 'billing_country_code', 'name_11172', 'email_953', 'linking_identity_395', 'periodic_311', 'ip_city', 'cc_v8', 'riskimal_class', 'ip_v3', 'cc_v6', 'linking_online_779', 'linking_online_745', 'prediction_value', 'email_73', 'linking_identity_72'}label columns diff {'label'}\", \"file_type\": null, \"id\": 661218, \"integration_type\": \"RECORD\", \"is_reviewed\": false, \"num_of_records\": \"80\", \"origin_url\": \"gs://superwise-oryandna-development/origin/year=2021/month=12/day=16/8000db7c-5e85-11ec-90b8-42013606f5d0.parquet\", \"status\": \"Failed\", \"model_id\": null, \"model_name\": null, \"transaction_id\": \"8000db7c-5e85-11ec-90b8-42013606f5d0\", \"version_id\": null}"
    get_response.status_code = 200
    monkeypatch.setattr(requests, "post", lambda *args, **kwargs: the_response)
    monkeypatch.setattr(requests, "get", lambda *args, **kwargs: get_response)


@pytest.fixture(scope="function")
def mock_azure_blob_reader(monkeypatch):
    monkeypatch.setattr(AzureClientReader, "get_file_from_blob_storage", lambda *args, **kwargs: dict(data="content"))


@pytest.fixture(scope="function")
def mock_boto_client(monkeypatch):
    class BotoClient:
        def __init__(self, *args, **kwargs):
            pass

        def get_object(self, *args, **kwargs):
            class ObjectResposne:
                def read(self):
                    return "blablabla"

            return dict(Body=ObjectResposne())

    monkeypatch.setattr(boto3, "client", lambda *args, **kwargs: BotoClient())


@pytest.fixture(scope="function")
def get_transaction_records_payload():
    with open(f"{PROJECT_ROOT}/tests/resources/transaction/records_payload.json") as f:
        return json.loads(f.read())


def test_transaction_records(mock_transaction_requests, sw, get_transaction_records_payload):
    status = sw.transaction.log_records(model_id=1, records=get_transaction_records_payload)
    assert isinstance(status, dict) and "transaction_id" in status.keys()
    status = sw.transaction.log_records(model_id="test", version_id=1, records=get_transaction_records_payload)
    assert isinstance(status, dict) and "transaction_id" in status.keys()


def test_transaction_file(mock_transaction_requests, sw):
    status = sw.transaction.log_file("gs://fvsdfvfdv")
    assert isinstance(status, dict) and "transaction_id" in status.keys()


def test_transaction_with_wrong_file_path(mock_transaction_requests, sw):
    with pytest.raises(SuperwiseValidationException):
        sw.transaction.log_file("wrong path")


def test_transaction_upload_from_gcs(mock_transaction_requests, mock_gcp_client, sw):
    status = sw.transaction.log_from_gcs(
        file_path="gs://superwise-oryan-test/new_integration_tests_binary_classification_predictions.csv",
        service_account={},
        metadata={"test": "value"},
    )

    assert isinstance(status, dict) and "transaction_id" in status.keys()


def test_transaction_upload_from_s3(mock_transaction_requests, mock_boto_client, sw):
    status = sw.transaction.log_from_s3(
        file_path="s3://superwise-oryan-test/new_integration_tests_binary_classification_predictions.csv",
        aws_access_key_id="test",
        aws_secret_access_key="test",
        metadata={"test": "value"},
    )
    assert isinstance(status, dict) and "transaction_id" in status.keys()


def test_transaction_upload_from_local_file(mock_transaction_requests, sw):
    status = sw.transaction.log_from_local_file(
        file_path=os.path.join(PROJECT_ROOT, "tests", "resources", "transaction", "local_records_file.json"),
        metadata={"test": "value"},
    )
    assert isinstance(status, dict) and "transaction_id" in status.keys()
    assert isinstance(status, dict) and "metadata" in status.keys()


def test_transaction_upload_from_azure(mock_transaction_requests, mock_azure_blob_reader, sw):
    status = sw.transaction.log_from_azure(
        azure_uri="abfs://container/blob", connection_string="connection string", metadata={"test": "value"}
    )

    assert isinstance(status, dict) and "transaction_id" in status.keys()
    assert isinstance(status, dict) and "metadata" in status.keys()


def test_transaction_upload_from_not_existing_local_file(sw):
    with pytest.raises(FileNotFoundError):
        sw.transaction.log_from_local_file("invalid_path")


def test_get_transaction(mock_transaction_requests, sw):
    transaction = sw.transaction.get(transaction_id="blabla")
    assert isinstance(transaction, Transaction)



""" This module implement Dataset model  """
from typing import Dict
from typing import List
from typing import Union

from superwise.models.base import BaseModel
from superwise.resources.superwise_enums import DatasetStatus
from superwise.resources.superwise_enums import DatasetType


class Dataset(BaseModel):
    """DataSet"""

    def __init__(
        self,
        name: str,
        files: Union[str, List[str]],
        project_id: int,
        type: Union[DatasetType, str] = DatasetType.TRAIN,
        dtypes: Dict[str, str] = None,
        roles: Dict[str, str] = None,
        **kwargs
    ):
        """
        ### Description:

        Constructor for Dataset class

        ### Args:

        `name`: The name of the dataset.

        `files`: The raw data of the dataset. Can be provided as local or cloud file paths. (GCS or S3)

        `project_id`: The ID of the project which this dataset will be assigned to.

        `type`: The type of the dataset (See 'superwise.resources.superwise_enums.DatasetType' enum). Default 'TRAIN'.

        `dtypes`: An optional mapping between columns and their dtypes.
                  if not provided, will be inferred.

        `roles`: An optional mapping between columns and their roles.
                 If not provided, will be inferred.
        """
        BaseModel.__init__(self, **kwargs)
        self.project_id = project_id
        self.name = name
        self.type = type.value if isinstance(type, DatasetType) else type
        self.dtypes = dtypes
        self.roles = roles
        self.files = files
        self.matching_entities = kwargs.get("matching_entities")
        self.internal_files = kwargs.get("internal_files")
        self.created_at = kwargs.get("created_at")
        self.created_by = kwargs.get("created_by")
        self.num_rows = kwargs.get("num_rows")
        self.num_columns = kwargs.get("num_columns")
        self.head = kwargs.get("head")
        self.status = kwargs.get("status", DatasetStatus.UNKNOWN.value)
        self.status_reason = kwargs.get("status_reason")
        self.is_static = kwargs.get("is_static", True)



from datetime import datetime

import requests
from google.auth import jwt
from google.cloud import storage
from google.oauth2.credentials import Credentials

from superwise import Config
from superwise.utils.client import Client
from superwise.utils.exceptions import SuperwiseStorageUploadGCSError
from superwise.utils.storage.gcs import GcsClient
from superwise.utils.storage.internal_storage.internal_storage import InternalStorage


class GCSInternalStorage(InternalStorage):
    def __init__(self, superwise_web_client: Client, bucket_name: str):
        super().__init__(superwise_web_client)
        self._bucket_name = bucket_name
        self._bucket = None
        self._gcs_client = GcsClient()

    def _put_in_storage(self, blob_name: str, content: str):
        return self._gcs_client.upload_string_to_bucket(content, blob_name, self._bucket)

    def _is_cloud_token_active(self) -> bool:
        expired_time = datetime.fromtimestamp(self._jwt_decoded["exp"])
        return datetime.utcnow() < expired_time

    def _refresh_cloud_token(self, file_url: str):
        super()._refresh_cloud_token(file_url)
        self._jwt_decoded = jwt.decode(self._cloud_token, verify=False)
        access_token = self._exchange_jwt_for_access_token()
        credentials = Credentials(token=access_token)
        gcs_client = storage.Client(credentials=credentials, project=Config.ENVIRONMENT)
        self._bucket = gcs_client.bucket(self._bucket_name)

    def _exchange_jwt_for_access_token(self):
        """
        This function takes a Signed JWT and exchanges it for a Google OAuth Access Token
        """
        params = {"grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer", "assertion": self._cloud_token}
        resp = requests.post(self._jwt_decoded["aud"], data=params)

        if resp.ok:
            return resp.json()["access_token"]

        raise SuperwiseStorageUploadGCSError(f"Failed to authenticate to google - {resp.text}")



""" wrapper for requests class and superwise token handling """
import json

import jwt
import requests
from requests import Response

from superwise import logger
from superwise.config import Config
from superwise.models.validation_error import ValidationError
from superwise.utils.exceptions import *


def token_retry(input_func):
    """
    ### Description:

    A decorator for token retry handling (ie, refresh token if needed)

    """

    def wrapper(*args, **kwargs) -> Response:
        try:
            res = input_func(*args, **kwargs)
        except SuperwiseAuthException:
            args[0].refresh_token()
            res = input_func(*args, **kwargs)
        return res

    return wrapper


class Client:
    """A Client for http requests, set of wrappers around requests library"""

    def __init__(self, client_id, secret, api_host, email=None, password=None):
        """

        ### Args:

        `client_id`:  user client token id (string)

        `secret`: secret token (string)

        `api_host`:  superwise api server host

        `email`: optional email

        `password`: Optional password

        """
        self.client_id = client_id
        self.secret = secret
        self.api_host = api_host
        self.email = email
        self.password = password
        self._accessToken = None
        self._accessToken = self.get_access_token()
        self._refreshToken = None
        self._expires = None
        self._expiresIn = None
        self.tenant_id = self.get_tenant_id(self.accessToken)
        self.logger = logger
        self.headers = self.build_headers()

    def get_access_token(self):
        if self._accessToken is not None:
            return self._accessToken
        return self.get_token().get("accessToken", None)

    def get_tenant_id(self, token) -> str:
        """
        ### Description:

        Get tenant id from JWT token

        ### Args:

        `token`: secret token (string)

        ### Return:

        tenant_id string
        """
        try:
            jwt_payload = jwt.decode(token, "secret", options={"verify_signature": False})
        except:
            raise SuperwiseTokenException("error parsing jwt token: {}".format(token))
        if jwt_payload:
            return jwt_payload.get("tenantId")
        else:
            raise SuperwiseException("error decoding JWT")

    def build_headers(self):
        """
        ### Description:

        ### Return:
        Headers to send (dict)
        """
        return {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": "Bearer " + self.accessToken,
        }

    # TODO: actually refresh token
    def refresh_token(self):
        """
        ### Description:

        refresh  bearer token

        ### Return:

        token string
        """
        return self.get_token()

    def get_token(self) -> str:
        """
        ### Description:
        Get bearer token to use in each API call

        ### Return:

        token string
        """

        headers = {"Accept": "application/json", "Content-Type": "application/json"}
        if self.email:
            url = "{}/identity/resources/auth/v1/user".format(Config.AUTH_URL)
            params = {"email": self.email, "password": self.password}
        else:
            params = {"clientId": self.client_id, "secret": self.secret}
            url = "{}/identity/resources/auth/v1/api-token".format(Config.AUTH_URL)
        res = self._post(url, params=params, headers=headers)

        error = False
        try:
            res = res.json()
            self._accessToken = res.get("accessToken", None)
            self._refreshToken = res.get("refreshToken", None)
            self._expires = res.get("expires", None)
            self._expiresIn = res.get("_expiresIn", None)
        except:
            error = True
        if error:
            raise SuperwiseAuthException("Error get or refresh token url={} params={} ".format(url, params))

        return res

    def _post(self, url, params, headers=None) -> Response:
        """
        ### Description:

        Wrapper to requests.post(), with no refresh token handling

        ### Args:

        `url`: url string

        `params`: json paramters to send

        `headers`: headers string

        ### Return:

        Response object (requests package)
        """
        headers = self.headers if not headers else headers
        logger.debug("POST:  {} params: {}".format(url, params))
        res = requests.post(url, json=params, headers=headers)
        return self._check_res(res)

    @token_retry
    def post(self, url, params, headers=None) -> Response:
        """
        ### Description:

        Wrapper to requests.post(), with no refresh token handling

        ### Args:

        `url`: url string

        `params`: json paramters to send

        `headers`: headers string

        ### Return:

        Response object (requests package)
        """
        return self._post(url, params, headers)

    def _check_res(self, res) -> Response:
        """
        ### Description:

        Check response status and raise Exception if error

        ### Args:

        `res`: response object from requests call

        ### Return:

        response object
        """
        if res.status_code == 422:
            try:
                body = json.loads(res.content)
            except Exception as ex:
                raise Exception("error loading json, status code {} text {}".format(res.status_code, res.content))

            ValidationError(http_status_code=res.status_code, http_error_reason=res.reason, body=body)
        if res.status_code in [401, 403]:
            logger.error(f"error {res.status_code} - {res.content}")
            raise SuperwiseAuthException(res.content)
        if res.status_code in [500]:
            logger.error(f"error 500 {res.content}")
            raise SuperWiseInternalServerError(res.content)
        if res.status_code not in [200, 201, 202, 204]:
            logger.error(f"error {res.status_code} - {res.content}")
            raise Exception(res.content)
        return res

    @token_retry
    def get(self, url, query_params=None, headers=None, timeout=None) -> Response:
        """
        ### Description:

        Wrapper to requests.get(), with no refresh token handling

        ### Args:

        `url`: url string

        `query_params`: dictionary of parameters to add as query string

        `headers`: headers dictionary

        ### Return:

        Response object (requests package)
        """
        headers = self.headers if not headers else headers
        logger.debug("GET {} query params: {}".format(url, query_params))
        res = requests.get(url=url, headers=headers, params=query_params, timeout=timeout)
        return self._check_res(res)

    @token_retry
    def delete(self, url, headers=None) -> Response:
        """
        Wrapper for reuqests.delete(),

        :param idx: id int
        :return: Response object (requests package)
        """

        """
        ### Description:

        Wrapper for reuqests.delete()

        ### Args:

        `url`: url of resource to delete

        ### Return:

        Response object (requests package)
        """

        headers = self.headers if not headers else headers
        logger.debug("DELETE {}".format(url))
        res = requests.delete(url=url, headers=headers)
        return self._check_res(res)

    @token_retry
    def patch(self, url, params, headers=None) -> Response:
        """
        ### Description:

        Wrapper to requests.patch(), with no refresh token handling

        ### Args:

        `url`: url string

        `params`: dictionary of fields to update

        `headers`: headers dictionary

        ### Return:

        Response object (requests package)
        """

        headers = self.headers if not headers else headers
        logger.debug("PATCH {}  params: {} ".format(url, params))
        res = requests.patch(url, json=params, headers=headers)
        return self._check_res(res)

    def build_url(self, path):
        """
        ### Description:

        Build an url for a given path

        ### Args:

        `path`:  relative path, normally declared in each model (subclasses of this class)

        ### Return:

        URL string
        """
        return "https://{}/{}/{}".format(self.api_host, self.tenant_id, path)

    accessToken = property(get_access_token)



""" This module implement tasks functionality  """
from superwise.controller.base import BaseController
from superwise.models.model import Model
from superwise.models.project import Project


class ModelController(BaseController):
    """Class ModelController - responsible for task functionality"""

    def __init__(self, client, sw):
        """

        ### Args:

        `client`: superwise client object

        `sw`: superwise  object

        """
        super().__init__(client, sw)
        self.path = "admin/v1/models"
        self.model_name = "Model"

    def archive(self, model):
        """

        ### Args:

        `model`: model object to archive

        ### Return:

        response object
        """
        res = self.update(model.id, params={"is_archived": True})
        return res



import pytest
import requests
from pandas import DataFrame
from requests import Response

from superwise.models.notification import Notification
from superwise.resources.superwise_enums import NotifyUpon
from superwise.resources.superwise_enums import ScheduleCron


@pytest.fixture(scope="function")
def mock_monitor_get_policy_templates(monkeypatch):
    def mock_monitors_get_templates_logic(*args, **kwargs):
        get_response = Response()
        get_response._content = b"""
            [
          {
            "description": "Scanning the X top important features for long-lasting drift on the entire set level",
            "name": "Feature stability"
          },
          {
            "description": "Scan input drift on the segment level",
            "name": "Dataset shift"
          },
          {
           "description": "Alert when input is above a certain threshold from the baseline",
           "name": "Training-Serving skew"
          },
          {
            "description": "Missing values on a feature level on a segment level",
            "name": "Missing values"
          },
          {
            "description": "Anomaly in % of outliers on a feature level (numeric only)",
            "name": "Out-of-Range"
          },
          {
            "description": "Anomaly in % of new values on a feature level (categorical only)",
            "name": "New values"
          }]
            """
        get_response.status_code = 200
        return get_response

    monkeypatch.setattr(requests, "get", mock_monitors_get_templates_logic)


@pytest.fixture(scope="function")
def mock_empty_monitor_get_policy_templates(monkeypatch):
    def mock_monitors_get_templates_not_templates_logic(*args, **kwargs):
        get_response = Response()
        get_response._content = b""
        get_response.status_code = 200
        return get_response

    monkeypatch.setattr(requests, "get", mock_monitors_get_templates_not_templates_logic)


@pytest.fixture(scope="function")
def mock_bad_response_monitor_get_policy_templates(monkeypatch):
    def mock_monitors_get_templates_not_bad_status_code_logic(*args, **kwargs):
        get_response = Response()
        get_response._content = b"Internal Server Error"
        get_response.status_code = 500
        return get_response

    monkeypatch.setattr(requests, "get", mock_monitors_get_templates_not_bad_status_code_logic)


@pytest.fixture(scope="function")
def mock_monitor_post_create_policy_from_templates(monkeypatch):
    def mock_monitor_create_policy_from_template_logic(*args, **kwargs):
        post_response = Response()
        post_response.status_code = 201
        return post_response

    monkeypatch.setattr(requests, "post", mock_monitor_create_policy_from_template_logic)


def test_get_policy_templates(mock_monitor_get_policy_templates, sw):
    templates = sw.policy.get_policy_templates()
    assert isinstance(templates, DataFrame)
    assert sorted(templates.columns.tolist()) == ["description", "name"]
    assert len(templates) == 6


def test_get_policy_templates_when_no_template(mock_empty_monitor_get_policy_templates, sw):
    with pytest.raises(AssertionError):
        sw.policy.get_policy_templates()


def test_get_policy_templates_when_get_error_response(mock_bad_response_monitor_get_policy_templates, sw):
    with pytest.raises(Exception):
        sw.policy.get_policy_templates()


def test_create_policy_from_template(mock_monitor_post_create_policy_from_templates, sw):
    response = sw.policy.create_policy_from_template(
        model_id=1,
        template_name="template_id",
        policy_name="test policy",
        notification_channels=[Notification(id=1)],
        notify_upon=NotifyUpon.detection_and_resolution,
        schedule=ScheduleCron.EVERY_2ND_MONTH,
    )
    assert response.status_code == 201



from datetime import datetime
from http import HTTPStatus
from urllib.parse import urlparse

import requests
from azure.storage.blob import BlobClient

from superwise.utils.exceptions import SuperwiseRefreshCloudTokenError
from superwise.utils.exceptions import SuperwiseStorageUploadAzureError
from superwise.utils.storage import validate_file_size


class AzureClientUploader:
    url = "https://{account_name}.blob.core.windows.net/{container_name}/{blob_name}"
    internal_services_valid_url = "abfs://{container_name}/{blob_name}"

    def __init__(self, account_name, container_name):
        self._account_name = account_name
        self._container_name = container_name

    def upload(self, content, blob_name, sas_token):
        url = self.url.format(account_name=self._account_name, container_name=self._container_name, blob_name=blob_name)

        headers = {"x-ms-date": datetime.utcnow().strftime("%a %B %d %T %Y"), "x-ms-blob-type": "BlockBlob"}
        response = requests.put(url, data=content, headers=headers, params=sas_token)
        if response.status_code == HTTPStatus.FORBIDDEN:
            raise SuperwiseRefreshCloudTokenError(
                f"Forbidden access to azure's container {self._container_name}, " f"try to refresh the sas token"
            )
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            raise SuperwiseStorageUploadAzureError(f"Failed upload file to superwise storage {blob_name} with ext {e}")
        else:
            return self.internal_services_valid_url.format(container_name=self._container_name, blob_name=blob_name)


class AzureClientReader:
    def __init__(self, connection_string):
        self.connection_string = connection_string

    def get_file_from_blob_storage(self, file_path, validate_size=False) -> dict:
        container = self._extract_container_name_from_path(file_path)
        blob_name = self._extract_blob_name_from_path(file_path)
        blob_client = BlobClient.from_connection_string(
            conn_str=self.connection_string, container_name=container, blob_name=blob_name
        )
        size = None
        if validate_size:
            size = blob_client.get_blob_properties()["size"]
            validate_file_size(file_path, size)

        data = blob_client.download_blob().readall()
        return dict(filename=blob_name, data=data, size=size)

    @staticmethod
    def _extract_container_name_from_path(file_path):
        """
        'abfs://my_container/my/blob' --> 'my_container'
        """
        parsed_path = urlparse(file_path)
        container = parsed_path.hostname

        return container

    @staticmethod
    def _extract_blob_name_from_path(file_path):
        """
        'abfs://my_container/my/blob' --> 'my/blob'
        """
        parsed_path = urlparse(file_path)
        # removing first slash(/)
        blob_name = parsed_path.path[1:]

        return blob_name



""" This module implement suprewise custom exceptions  """


class SuperWiseInternalServerError(Exception):
    """SuperWiseInternalServerError customn exception"""


class SuperWiseValidationError(Exception):
    """SuperWiseValidationError customn exception"""


class SuperwiseException(SuperWiseValidationError):
    """SuperwiseException customn exception"""


class SuperwiseTokenException(SuperWiseValidationError):
    """SuperwiseException customn exception"""


class SuperwiseValidationException(SuperWiseValidationError):
    """SuperwiseValidationException customn exception"""


class SuperwiseUnsupportedException(SuperWiseValidationError):
    """SuperwiseValidationException customn exception"""


class SuperwiseAuthException(SuperWiseValidationError):
    """SuperwiseAuthException customn exception"""


class SuperwiseServiceAccountException(SuperWiseValidationError):
    """SuperwiseAuthException custom exception"""


class SuperwiseStorageDownloadS3Error(SuperWiseValidationError):
    """SuperwiseStorageDownloadS3Error custom exception"""


class SuperwiseStorageDownloadGCSError(SuperWiseValidationError):
    """SuperwiseStorageDownloadGCSError custom exception"""


class SuperwiseStorageUploadGCSError(SuperWiseValidationError):
    """SuperwiseStorageUploadGCSError custom exception"""


class SuperwiseStorageUploadAWSError(SuperWiseValidationError):
    """SuperwiseStorageUploadGCSError custom exception"""


class SuperwiseStorageUploadAzureError(SuperWiseValidationError):
    """SuperwiseStorageUploadAzureError custom exception"""


class SuperwiseRefreshCloudTokenError(Exception):
    """SuperwiseRefreshCloudTokenError custom exception"""


class SuperwiseTimeoutError(Exception):
    """Superwise timeout exception"""


class SuperwiseDatasetFailureError(Exception):
    """Superwise dataset creation error"""


class SuperwiseFileTooLargeError(Exception):
    """Superwise file too large error"""



""" This module implement s3 client functionality  """
from typing import Optional

import boto3

from superwise import logger
from superwise.utils.exceptions import SuperwiseException
from superwise.utils.exceptions import SuperwiseFileTooLargeError
from superwise.utils.exceptions import SuperwiseStorageDownloadS3Error
from superwise.utils.singelton import singleton
from superwise.utils.storage import extract_directory
from superwise.utils.storage import validate_file_size


@singleton
class S3Client:
    def __init__(self):
        self.logger = logger

    def get_s3_file(
        self,
        file_path: str,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        role_arn: Optional[str] = None,
        validate_size: bool = False,
    ):
        """
        ### Description:
        get a given s3 file from s3

        ### Args:
        `file_path`:  file path string

        `aws_access_key_id`:  aws access key id string

        `aws_secret_access_key`:  aws secret key  string

        `role_arn`:  aws role arn string

        ### Return:
        a tuple the s3 file key (path) and the content of the file
        """
        if not str(file_path).startswith("s3://"):
            self.logger.error(f"Failed upload file to superwise storage {file_path}")
            raise Exception("file_path must start with 's3://'")
        try:
            self.logger.info("Download file {} from s3".format(file_path))
            bucket, key = extract_directory(file_path)
            s3_client = self._create_s3_client(
                aws_access_key_id=aws_access_key_id, aws_secret_access_key=aws_secret_access_key, role_arn=role_arn
            )
            size = self._validate_object_size(file_path, s3_client) if validate_size else None
            resp = s3_client.get_object(Bucket=bucket, Key=key)
            return dict(filename=key, data=resp["Body"].read(), size=size)
        except SuperwiseFileTooLargeError:
            raise
        except Exception as e:
            self.logger.error(f"Error download file from customer s3 bucket {file_path} with ext {e}")
            raise SuperwiseStorageDownloadS3Error(f"Error download file from customer s3 bucket {file_path}")

    def _create_s3_client(self, aws_access_key_id: str = None, aws_secret_access_key: str = None, role_arn: str = None):
        """
        ### Description:
        create s3   client to work with

        ### Args:
        `aws_access_key_id`:  aws access key id string

        `aws_secret_access_key`:  aws secret key  string

        `role_arn`:  aws role arn string

        ### Return:
        boto3.client("s3") object
        """
        try:
            if role_arn is not None:
                self.logger.debug("Create S3 client from role_arn")
                sts_client = boto3.client("sts")
                assumed_role_object = sts_client.assume_role(RoleArn=role_arn, RoleSessionName="superwise-session")
                credentials = assumed_role_object["Credentials"]

                return boto3.client(
                    "s3",
                    aws_access_key_id=credentials["AccessKeyId"],
                    aws_secret_access_key=credentials["SecretAccessKey"],
                    aws_session_token=credentials["SessionToken"],
                )
            elif aws_access_key_id is not None and aws_secret_access_key is not None:
                self.logger.debug("Create S3 client with aws access key and secret key")
                return boto3.client(
                    "s3", aws_access_key_id=aws_access_key_id, aws_secret_access_key=aws_secret_access_key
                )

            else:
                self.logger.debug("Create s3 client with out any params")
                return boto3.client("s3")
        except Exception as e:
            self.logger.error(f"Error create s3 client, ext {e}")
            raise SuperwiseException("Error create s3 client - role arn or access and secret keys not provided")

    @staticmethod
    def _validate_object_size(file_url: str, client):
        logger.debug(f"Get metadata from S3 for file_url: {file_url}")
        bucket_name, object_key = extract_directory(file_url)
        metadata = client.head_object(Bucket=bucket_name, Key=object_key)
        size = metadata["ContentLength"]

        validate_file_size(file_path=file_url, file_size=size)
        return size



import abc
import logging
import uuid
from enum import Enum

from superwise import Config
from superwise.utils.client import Client
from superwise.utils.exceptions import SuperWiseValidationError

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class CloudProvider(Enum):
    GCP = "GCP"
    AZURE = "AZURE"
    AWS = "AWS"


class InternalStorage(abc.ABC):
    @staticmethod
    def generate_file_name(extension: str, prefix: str = None):
        file_prefix = ""
        if prefix:
            file_prefix = f"{prefix}_"
        return f"{file_prefix}{uuid.uuid4()}{extension}"

    def __init__(self, superwise_web_client: Client):
        self._superwise_web_client = superwise_web_client
        self._cloud_token = None

    def _refresh_cloud_token(self, file_url: str):
        url_path = f"admin/v1/settings/jwt-token?file_url={file_url}"
        resp = self._superwise_web_client.get(self._superwise_web_client.build_url(url_path))
        resp.raise_for_status()
        self._cloud_token = resp.json()

    @abc.abstractmethod
    def _is_cloud_token_active(self) -> bool:
        pass

    @abc.abstractmethod
    def _put_in_storage(self, blob_name: str, content: str):
        pass

    def upload(self, blob_path: str, content: str):
        """

        ### Description:

        Upload a file's content to Superwise's internal blob storage

        ### Args:

        `blob_path`: file path to upload to

        `content`: the content of the file

        """
        logger.debug(f"Upload file to Superwise's internal blob storage {blob_path}")
        if not self._cloud_token or not self._is_cloud_token_active():
            self._refresh_cloud_token(file_url=blob_path)
        try:
            file_url = self._put_in_storage(blob_path, content)
        except SuperWiseValidationError:
            raise
        except Exception as e:
            raise Exception(f"failed to upload blob to internal bucket - {e}")
        else:
            return file_url



""" This module implement notification functionality  """
from typing import Optional

from superwise.controller.base import BaseController
from superwise.models.notification import Notification
from superwise.resources.superwise_enums import NotificationType


class NotificationController(BaseController):
    """Class NotificationController - responsible for notification functionality"""

    def __init__(self, client, sw):
        """

        ### Args:

        `client`: superwise client object

        `sw`: superwise  object

        """
        super().__init__(client, sw)
        self.path = "notification/v1/notifications"
        self.model_name = "Notification"

    def create_email_notification(self, name: str, email: str):
        """

        ### Args:

        `name`: notification name

        `email`: notification email

        """
        notification_model: Notification = Notification(
            name=name, notification_metadata={"target": [email]}, type=NotificationType.Email.value
        )
        return super().create(notification_model)

    def create_slack_notification(self, name: str, webhook_url: str):
        """
        Create Slack webhook notification
                ### Args:

                `name`: notification name

                `webhook_url`: notification email

        """
        notification_model: Notification = Notification(
            name=name, notification_metadata={"target": webhook_url}, type=NotificationType.SlackWebhook.value
        )
        return super().create(notification_model)

    def create_pagerduty_notification(self, name: str, token: str):
        """
        Create PagerDuty notification
        ### Args:

        `name`: notification name

        `token`: PagerDuty integration IGcsTokenSet to notify on new incidents

        """
        notification_model: Notification = Notification(
            name=name, notification_metadata={"target": token}, type=NotificationType.PagerDuty.value
        )
        return super().create(notification_model)

    def create_webhook_notification(self, name: str, webhook_url: str):
        """
        Create Webhook notification
        ### Args:

        `name`: notification name

        `webhook_url`: webhook url

        """
        notification_model: Notification = Notification(
            name=name, notification_metadata={"target": webhook_url}, type=NotificationType.Webhook.value
        )
        return super().create(notification_model)

    def create_datadog_notification(self, name: str, dd_api_key: str, dd_application_key: str):
        """
        Create Datadog notification
        ### Args:

        `name`: notification name

        `dd_api_key`: API key will allow us to submit metrics and events to Datadog

        `dd_application_key`: Application keys will allows us to submit incidents to Datadog

        """
        notification_metadata = {"target": {"dd_api_key": dd_api_key, "dd_application_key": dd_application_key}}

        notification_model: Notification = Notification(
            name=name, notification_metadata=notification_metadata, type=NotificationType.Datadog.value
        )
        return super().create(notification_model)

    def get(self, name: Optional[str] = None):
        """

        ### Args:

        `name`: notification name

        """
        if name is not None:
            return super().get({"name": name})
        return super().get({})



import requests

from superwise.utils.client import Client
from superwise.utils.exceptions import SuperwiseStorageUploadAWSError
from superwise.utils.storage.internal_storage.internal_storage import InternalStorage


class AWSInternalStorage(InternalStorage):
    def __init__(self, superwise_web_client: Client, bucket_name: str):
        super().__init__(superwise_web_client)
        self._bucket_name = bucket_name

    def _put_in_storage(self, blob_name: str, content: str):
        try:
            url, fields = self._cloud_token["url"], self._cloud_token["fields"]
            file = {"file": (blob_name, content)}
            res = requests.post(url, data=fields, files=file)
            res.raise_for_status()
            return f"s3://{self._bucket_name}/{blob_name}"
        except Exception as e:
            raise SuperwiseStorageUploadAWSError(f"error upload file {blob_name} to internal bucket with ext{e}")

    def _is_cloud_token_active(self) -> bool:
        """
        cloud token is enable per file path to every upload a new token should be created
        """
        return False



""" This module implement version  model  """
import pandas as pd

from superwise.models.base import BaseModel


class Version(BaseModel):
    """Version model class, model  for version data"""

    def __init__(
        self,
        id=None,
        model_id=None,
        name=None,
        created_at=None,
        data_entities=None,
        baseline_df=None,
        status=None,
        dataset_id=None,
        **kwargs
    ):
        """
        ### Description:

        Constructor for Version class

        ### Args:

        `id`: id of Version

        `model_id`: id of model (model))

        `name`: name of version

        `created_at`:

        `data_entities`: list of DataEntity (List[DataEntity])

        `baseline_df`: pandas df of basline data

        `status`:

        `dataset_id`: Dataset to connect to this version
        """
        self.id = id
        self.model_id = model_id
        self.dataset_id = dataset_id
        self.name = name
        self.baseline_files = []
        self.created_at = created_at
        self.data_entities = data_entities or []
        self.status = status
        self.baseline_df = baseline_df



from enum import Enum


class DataEntityRole(Enum):
    """Data Entity Role enum"""

    ID = "id"
    TIMESTAMP = "time stamp"
    FEATURE = "feature"
    PREDICTION_PROBABILITY = "prediction probability"
    PREDICTION_VALUE = "prediction value"
    LABEL = "label"
    LABEL_TIMESTAMP = "label time stamp"
    LABEL_WEIGHT = "label weight"
    METADATA = "metadata"


class ModelTypes(Enum):
    """Data Types enum"""

    BINARY_CLASSIFICATION = "Binary Classification"
    BINARY_ESTIMATION = "Binary Estimation"
    REGRESSION = "Regression"
    MULTICLASS_CLASSIFICATION = "Multiclass Classification"


class FeatureType(Enum):
    """Feature Type enum"""

    NUMERIC = "Numeric"
    BOOLEAN = "Boolean"
    CATEGORICAL = "Categorical"
    TIMESTAMP = "Timestamp"
    UNKNOWN = "Unknown"


class CategoricalSecondaryType(Enum):
    """Categorical Secondary Type enum"""

    CONSTANT = "Cat_constant"
    DENSE = "Cat_dense"
    SPARSE = "Cat_sparse"


class NumericSecondaryType(Enum):
    """Numeric Secondary Type enum"""

    NUM_RIGHT_TAIL = "Num_right_tail"
    NUM_LEFT_TAIL = "Num_left_tail"
    NUM_CENTERED = "Num_centered"


class BooleanSecondaryType(Enum):
    """Boolean Secondary Type enum"""

    FLAG = "Boolean_flag"
    NUMERIC = "Boolean_numeric"


def get_enum_value(v):
    """
    ### Description:

    This function  enum property and return the value of the enum

    ### Args:

    `v`:  an enum object

    """
    if isinstance(v, Enum):
        return v.value
    else:
        return v


class NotificationType(Enum):
    """Notification type Type enum"""

    SlackWebhook = "SlackWebhook"
    Webhook = "Webhook"
    PagerDuty = "PagerDuty"
    Email = "Email"
    NewRelic = "NewRelic"
    Datadog = "Datadog"


class ScheduleCron(Enum):
    EVERY_DAY_AT_1AM = "0 01 * * *"
    EVERY_DAY_AT_2AM = "0 02 * * *"
    EVERY_DAY_AT_3AM = "0 03 * * *"
    EVERY_DAY_AT_4AM = "0 04 * * *"
    EVERY_DAY_AT_5AM = "0 05 * * *"
    EVERY_DAY_AT_6AM = "0 06 * * *"
    EVERY_DAY_AT_7AM = "0 07 * * *"
    EVERY_DAY_AT_8AM = "0 08 * * *"
    EVERY_DAY_AT_9AM = "0 09 * * *"
    EVERY_DAY_AT_10AM = "0 10 * * *"
    EVERY_DAY_AT_11AM = "0 11 * * *"
    EVERY_DAY_AT_NOON = "0 12 * * *"
    EVERY_DAY_AT_1PM = "0 13 * * *"
    EVERY_DAY_AT_2PM = "0 14 * * *"
    EVERY_DAY_AT_3PM = "0 15 * * *"
    EVERY_DAY_AT_4PM = "0 16 * * *"
    EVERY_DAY_AT_5PM = "0 17 * * *"
    EVERY_DAY_AT_6PM = "0 18 * * *"
    EVERY_DAY_AT_7PM = "0 19 * * *"
    EVERY_DAY_AT_8PM = "0 20 * * *"
    EVERY_DAY_AT_9PM = "0 21 * * *"
    EVERY_DAY_AT_10PM = "0 22 * * *"
    EVERY_DAY_AT_11PM = "0 23 * * *"
    EVERY_DAY_AT_MIDNIGHT = "0 0 * * *"
    EVERY_WEEK = "0 0 * * 0"
    EVERY_WEEKDAY = "0 0 * * 1-5"
    EVERY_WEEKEND = "0 0 * * 6,0"
    EVERY_1ST_DAY_OF_MONTH_AT_MIDNIGHT = "0 0 1 * *"
    EVERY_1ST_DAY_OF_MONTH_AT_NOON = "0 12 1 * *"
    EVERY_2ND_MONTH = "0 0 1 */2 *"
    EVERY_QUARTER = "0 0 1 */3 *"
    EVERY_6_MONTHS = "0 0 1 */6 *"
    EVERY_YEAR = "0 0 1 1 *"
    MONDAY_TO_FRIDAY_AT_1AM = "0 0 01 * * 1-5"
    MONDAY_TO_FRIDAY_AT_2AM = "0 0 02 * * 1-5"
    MONDAY_TO_FRIDAY_AT_3AM = "0 0 03 * * 1-5"
    MONDAY_TO_FRIDAY_AT_4AM = "0 0 04 * * 1-5"
    MONDAY_TO_FRIDAY_AT_5AM = "0 0 05 * * 1-5"
    MONDAY_TO_FRIDAY_AT_6AM = "0 0 06 * * 1-5"
    MONDAY_TO_FRIDAY_AT_7AM = "0 0 07 * * 1-5"
    MONDAY_TO_FRIDAY_AT_8AM = "0 0 08 * * 1-5"
    MONDAY_TO_FRIDAY_AT_9AM = "0 0 09 * * 1-5"
    MONDAY_TO_FRIDAY_AT_09_30AM = "0 30 09 * * 1-5"
    MONDAY_TO_FRIDAY_AT_10AM = "0 0 10 * * 1-5"
    MONDAY_TO_FRIDAY_AT_11AM = "0 0 11 * * 1-5"
    MONDAY_TO_FRIDAY_AT_11_30AM = "0 30 11 * * 1-5"
    MONDAY_TO_FRIDAY_AT_12PM = "0 0 12 * * 1-5"
    MONDAY_TO_FRIDAY_AT_1PM = "0 0 13 * * 1-5"
    MONDAY_TO_FRIDAY_AT_2PM = "0 0 14 * * 1-5"
    MONDAY_TO_FRIDAY_AT_3PM = "0 0 15 * * 1-5"
    MONDAY_TO_FRIDAY_AT_4PM = "0 0 16 * * 1-5"
    MONDAY_TO_FRIDAY_AT_5PM = "0 0 17 * * 1-5"
    MONDAY_TO_FRIDAY_AT_6PM = "0 0 18 * * 1-5"
    MONDAY_TO_FRIDAY_AT_7PM = "0 0 19 * * 1-5"
    MONDAY_TO_FRIDAY_AT_8PM = "0 0 20 * * 1-5"
    MONDAY_TO_FRIDAY_AT_9PM = "0 0 21 * * 1-5"
    MONDAY_TO_FRIDAY_AT_10PM = "0 0 22 * * 1-5"
    MONDAY_TO_FRIDAY_AT_11PM = "0 0 23 * * 1-5"


class NotifyUpon(Enum):
    detection = "detection"
    resolution = "resolution"
    detection_and_resolution = "detection_and_resolution"


class SegmentStatus(Enum):
    __name__ = "SegmentStatus"
    ACTIVE = "ACTIVE"
    ERROR = "ERROR"
    PENDING = "PENDING"
    ARCHIVED = "ARCHIVED"


class SegmentCondition(Enum):
    __name__ = "SegmentCondition"
    GREATER_THAN_EQ = ">="
    LESS_THAN_EQ = "<="
    GREATER_THAN = ">"
    LESS_THAN = "<"
    IS_NULL = "is null"
    IN = "in"
    NOT_IN = "not in"
    EQUALS = "=="
    BETWEEN = "between"


class DatasetType(Enum):
    TRAIN = "Train"
    TEST = "Test"
    VALIDATION = "Validation"
    OTHER = "Other"


class DatasetStatus(Enum):
    FILE_UPLOADED = "file_uploaded"
    INFERRING_DTYPES = "inferring_datatypes"
    INFERRED_DTYPES = "inferred_dtype"
    VALIDATING_DTYPES = "validating_datatypes"
    VALIDATED_DTYPES = "validated_datatypes"
    INFERRING_ROLES = "inferring_roles"
    INFERRED_ROLES = "inferred_roles"
    VALIDATING_ROLES = "validating_roles"
    VALIDATED_ROLES = "validated_roles"
    MATCHING_ENTITIES = "matching_entities"
    ENTITIES_MATCHED = "entities_matched"
    SUMMARIZING = "summarizing"
    SUMMARIZED = "summarized"
    VERSION_CREATED = "version_created"
    UNKNOWN = "unknown"
    FAILED = "failed"
    FULL_FLOW_PROCESSING = "processing"


class DatasetOnFailureOptions(Enum):
    IGNORE = "ignore"
    RAISE = "raise"



import pytest
import requests
from requests import Response

from superwise.models.notification import Notification


@pytest.fixture(scope="function")
def mock_notification_requests(monkeypatch):
    get_response = Response()
    get_response._content = b'{"id": 14, "name": "oryan test 123", "type": "Email", "is_enabled": true, "is_hide": false, "created_at": 1647765197, "notification_metadata": {"target": ["oryan.omer@superwise.ai"]}, "last_delivery_status": "Success", "last_delivery_ts": 1647765208}'
    get_response.status_code = 201
    monkeypatch.setattr(requests, "get", lambda *args, **kwargs: get_response)
    monkeypatch.setattr(requests, "post", lambda *args, **kwargs: get_response)


@pytest.fixture(scope="function")
def mock_slack_notification_requests(monkeypatch):
    get_response = Response()
    get_response._content = b'{"id": 14, "name": "oryan test 123", "type": "SlackWebhook", "is_enabled": true, "is_hide": false, "created_at": 1647765197, "notification_metadata": {"target": "https://hooks.slack.com/services/T00000000/B00000000/XXXXXXXXXXXXXXXXXXXXXXXX"}, "last_delivery_status": "Success", "last_delivery_ts": 1647765208}'
    get_response.status_code = 201
    monkeypatch.setattr(requests, "get", lambda *args, **kwargs: get_response)
    monkeypatch.setattr(requests, "post", lambda *args, **kwargs: get_response)


def test_create_email_notification(mock_notification_requests, sw):
    notification = sw.notification.create_email_notification(name="oryan test 123", email="oryan.omer@superwise.ai")
    assert isinstance(notification, Notification)
    assert notification.name == "oryan test 123"


def test_create_slack_notification(mock_slack_notification_requests, sw):
    notification = sw.notification.create_slack_notification(
        name="oryan test 123",
        webhook_url="https://hooks.slack.com/services/T00000000/B00000000/XXXXXXXXXXXXXXXXXXXXXXXX",
    )
    assert isinstance(notification, Notification)
    assert notification.name == "oryan test 123"


def test_create_webhook_notification(mock_notification_requests, sw):
    notification = sw.notification.create_webhook_notification(
        name="oryan test 123",
        webhook_url="https://hooks.slack.com/services/T00000000/B00000000/XXXXXXXXXXXXXXXXXXXXXXXX",
    )
    assert isinstance(notification, Notification)
    assert notification.name == "oryan test 123"


def test_create_datadog_notification(mock_notification_requests, sw):
    notification = sw.notification.create_datadog_notification(
        name="oryan test 123", dd_api_key="abc", dd_application_key="bcd"
    )
    assert isinstance(notification, Notification)
    assert notification.name == "oryan test 123"


def test_create_pagerduty_notification(mock_slack_notification_requests, sw):
    notification = sw.notification.create_pagerduty_notification(name="oryan test 123", token="test")
    assert isinstance(notification, Notification)
    assert notification.name == "oryan test 123"


def test_get_notification(mock_notification_requests, sw):
    notification = sw.notification.get(name="oryan test 123")
    assert isinstance(notification, Notification)



import boto3
import pytest
import requests
from _pytest.monkeypatch import MonkeyPatch

from project_root import PROJECT_ROOT
from superwise.models.dataset import Dataset
from superwise.utils.exceptions import SuperwiseDatasetFailureError
from superwise.utils.exceptions import SuperwiseFileTooLargeError
from superwise.utils.exceptions import SuperwiseTimeoutError
from tests.conftest import parametrized_mock_gcp_client

monkey_patch = MonkeyPatch()


@pytest.fixture
def mock_boto_client():
    parametrized_mock_boto_client(1000)


def parametrized_mock_boto_client(file_size):
    class BotoClient:
        def __init__(self, *args, **kwargs):
            pass

        def get_object(self, *args, **kwargs):
            class ObjectResposne:
                def read(self):
                    return "blablabla"

            return dict(Body=ObjectResposne())

        def head_object(self, *args, **kwargs):
            return {"ContentLength": file_size}

    monkey_patch.setattr(boto3, "client", lambda *args, **kwargs: BotoClient())


@pytest.fixture
def mock_post_response(monkeypatch):
    post_response = requests.Response()
    post_response._content = (
        b'{ "id" : "123" ,"status" : "summarized", "project_id": 1, "name": "test", "files": ' b'"test.csv"}'
    )
    post_response.status_code = 201
    monkeypatch.setattr(requests, "post", lambda *args, **kwargs: post_response)


@pytest.fixture
def mock_file_uploaded_post_response(monkeypatch):
    post_response = requests.Response()
    post_response._content = (
        b'{ "id" : "123" ,"status" : "file_uploaded", "project_id": 1, "name": "test", "files": ' b'"test.csv"}'
    )
    post_response.status_code = 201
    monkeypatch.setattr(requests, "post", lambda *args, **kwargs: post_response)


@pytest.fixture
def mock_get_response(monkeypatch):
    get_response = requests.Response()
    get_response._content = (
        b'{ "id" : "123" ,"status" : "summarized", "project_id": 1, "name": "test", "files": ' b'"test.csv"}'
    )
    get_response.status_code = 200
    monkeypatch.setattr(requests, "get", lambda *args, **kwargs: get_response)


@pytest.fixture
def mock_failed_get_response(monkeypatch):
    get_response = requests.Response()
    get_response._content = (
        b'{ "id" : "123" ,"status" : "failed", "status_reason": "internal_error", "project_id": 1, "name": "test", "files": '
        b'"test.csv"}'
    )
    get_response.status_code = 200
    monkeypatch.setattr(requests, "get", lambda *args, **kwargs: get_response)


@pytest.fixture
def mock_not_finished_get_response(monkeypatch):
    get_response = requests.Response()
    get_response._content = (
        b'{ "id" : "123" ,"status" : "inferring_roles", "project_id": 1, "name": "test", "files": ' b'"test.csv"}'
    )
    get_response.status_code = 200
    monkeypatch.setattr(requests, "get", lambda *args, **kwargs: get_response)


@pytest.fixture
def files():
    return ["gs://bucket/test.csv", "s3://bucket/test.csv", f"{PROJECT_ROOT}/tests/resources/dataset/test.csv"]


@pytest.fixture
def basic_dataset(files):
    return Dataset(name="test", files=files, project_id=1)


def test_create_dataset_default_params(
    sw, mock_gcp_client, files, basic_dataset, mock_boto_client, mock_get_response, mock_post_response
):
    new_dataset = sw.dataset.create(basic_dataset)

    assert new_dataset.status == "summarized"
    assert new_dataset.id == "123"
    assert new_dataset.project_id == 1
    assert new_dataset.name == "test"
    assert new_dataset.files == "test.csv"


def test_create_dataset_on_failure_raise(
    sw, mock_gcp_client, files, basic_dataset, mock_boto_client, mock_failed_get_response, mock_post_response
):
    with pytest.raises(SuperwiseDatasetFailureError) as e:
        sw.dataset.create(basic_dataset)

    assert (
        e.value.args[0] == f"Dataset creation failed for ID '123'. "
        f"got status: 'failed', with reason: 'internal_error'"
    )


def test_create_dataset_on_failure_ignore(
    sw, mock_gcp_client, files, basic_dataset, mock_boto_client, mock_failed_get_response, mock_post_response
):
    new_dataset = sw.dataset.create(basic_dataset, on_failure="ignore")

    assert new_dataset.status == "failed"
    assert new_dataset.id == "123"
    assert new_dataset.project_id == 1
    assert new_dataset.name == "test"
    assert new_dataset.files == "test.csv"


def test_create_dataset_do_not_wait_until_complete(
    sw, mock_gcp_client, files, basic_dataset, mock_boto_client, mock_get_response, mock_post_response
):
    new_dataset = sw.dataset.create(basic_dataset, wait_until_complete=False)

    assert new_dataset.status == "summarized"
    assert new_dataset.id == "123"
    assert new_dataset.project_id == 1
    assert new_dataset.name == "test"
    assert new_dataset.files == "test.csv"


def test_create_dataset_time_out(
    sw, mock_gcp_client, files, basic_dataset, mock_boto_client, mock_not_finished_get_response, mock_post_response
):
    with pytest.raises(SuperwiseTimeoutError) as e:
        sw.dataset.create(basic_dataset, timeout_seconds=1)

    assert (
        e.value.args[0] == "Timed out while waiting for dataset with ID: '123' to be in one of "
        f"the statuses ['summarized', 'failed'], "
        f"got status: 'inferring_roles', with reason: '<unknown reason>'"
    )


def test_create_dataset_on_timeout_ignore(
    sw,
    mock_gcp_client,
    files,
    basic_dataset,
    mock_boto_client,
    mock_not_finished_get_response,
    mock_file_uploaded_post_response,
):
    new_dataset = sw.dataset.create(basic_dataset, timeout_seconds=1, on_failure="ignore")

    assert new_dataset.status == "file_uploaded"
    assert new_dataset.id == "123"
    assert new_dataset.project_id == 1
    assert new_dataset.name == "test"
    assert new_dataset.files == "test.csv"


def test_create_dataset_single_file_too_large(
    sw, basic_dataset, mock_boto_client, mock_get_response, mock_post_response
):
    parametrized_mock_gcp_client(160000000)

    with pytest.raises(SuperwiseFileTooLargeError) as e:
        sw.dataset.create(basic_dataset)

    assert e.value.args[0] == "File 'gs://bucket/test.csv' is too large: (160.0 MB), maximum size allowed is 100.0 MB"


def test_create_dataset_total_size_too_large(sw, basic_dataset, mock_get_response, mock_post_response):
    parametrized_mock_gcp_client(100000000)
    parametrized_mock_boto_client(60000000)

    with pytest.raises(SuperwiseFileTooLargeError) as e:
        sw.dataset.create(basic_dataset)

    assert (
        e.value.args[0] == "The *total* size of the files: ['gs://bucket/test.csv', 's3://bucket/test.csv'] "
        "is too large: (160.0 MB), maximum size allowed is 100.0 MB"
    )



import pandas as pd

from superwise.controller.infer import _infer_entity_type
from superwise.controller.infer import infer_dtype


def test_infer_bool():
    srs = pd.Series([False, True], name="bool_feature")
    dtype = _infer_entity_type(srs)
    assert dtype == "Boolean"


def test_infer_numeric():
    srs = pd.Series([1, 2, 3, 4, 1000])
    dtype = _infer_entity_type(srs)
    assert dtype == "Numeric"


def test_infer_categorical():
    srs = pd.Series(["abc", "abcd", "abc", "adsfasfdasdf"])
    dtype = _infer_entity_type(srs)
    assert dtype == "Categorical"

    srs = pd.Series(["abc", 12, 12312312312, 1231231231, 12, 12, 12, 12])
    dtype = _infer_entity_type(srs)
    assert dtype == "Categorical"


def test_infer_timedate():
    srs = pd.Series([pd.Timestamp("2022-04-11 08:26:33+0100")])
    dtype = _infer_entity_type(srs)
    assert dtype == "Timestamp"


def test_infer_unknown_empty_series():
    srs = pd.Series([])
    dtype = _infer_entity_type(srs)
    assert dtype == "Unknown"


def test_infer_unknown_series():
    srs = pd.Series([None, None, None])
    dtype = _infer_entity_type(srs)
    assert dtype == "Unknown"


def test_infer_timestamp_pd_timestamp():
    srs = pd.Series([pd.Timestamp("01-01-2010"), pd.Timestamp("01-01-2011")])
    dtype = _infer_entity_type(srs)
    assert dtype == "Timestamp"


def test_infer_timestamp_from_str():
    srs = pd.Series(["01-01-2010", "01-01-2011"])
    dtype = _infer_entity_type(srs)
    assert dtype == "Timestamp"

    srs = pd.Series(["01-01-2010 12:12", "01-01-2011 11:11"])
    dtype = _infer_entity_type(srs)
    assert dtype == "Timestamp"


def test_infer_dtype():
    df = pd.DataFrame({"f_bool": [True, False, True], "f_cat": ["Linux", "Win", "IOS"]})
    ret = infer_dtype(df)
    assert ret == {"f_bool": "Boolean", "f_cat": "Categorical"}



from datetime import datetime
from urllib.parse import parse_qs

from superwise.utils.client import Client
from superwise.utils.exceptions import SuperwiseRefreshCloudTokenError
from superwise.utils.exceptions import SuperwiseStorageUploadAzureError
from superwise.utils.storage.azure import AzureClientUploader
from superwise.utils.storage.internal_storage.internal_storage import InternalStorage


class AzureInternalStorage(InternalStorage):
    url = "https://{account_name}.blob.core.windows.net/{container_name}/{blob_name}"

    def __init__(self, superwise_client: Client, account_name: str, container_name: str):
        super().__init__(superwise_client)
        self._account_name = account_name
        self._azure_client = AzureClientUploader(account_name, container_name)

    def _is_cloud_token_active(self) -> bool:
        parsed_token = parse_qs(self._cloud_token)
        se = parsed_token["se"][0]
        expired_time = datetime.strptime(se, "%Y-%m-%dT%H:%M:%SZ")
        return datetime.utcnow() < expired_time

    def _put_in_storage(self, blob_name: str, content: str):
        try:
            url = self._azure_client.upload(content, blob_name, self._cloud_token)
        except SuperwiseRefreshCloudTokenError:
            # token has expired between the time of checking token expiration and sending the actual request
            self._refresh_cloud_token()
            url = self._azure_client.upload(content, blob_name, self._cloud_token)
        except Exception as e:
            raise SuperwiseStorageUploadAzureError(f"Failed upload file to superwise storage {blob_name} with ext {e}")

        return url



""" This module implement singleton decorator  """


def singleton(cls):
    """
    Decorator implementation for decorator

    """
    instance = [None]

    def wrapper(*args, **kwargs):
        if instance[0] is None:
            instance[0] = cls(*args, **kwargs)
        return instance[0]

    return wrapper



import json

import jwt
import pandas as pd
import pytest
import requests
from requests import Response

from project_root import PROJECT_ROOT
from superwise.controller.infer import infer_dtype
from superwise.models.data_entity import DataEntity
from superwise.models.model import Model
from superwise.models.version import Version
from superwise.resources.superwise_enums import DataEntityRole
from superwise.resources.superwise_enums import FeatureType


@pytest.fixture(scope="function")
def mock_jwt_decode(monkeypatch):
    monkeypatch.setattr(jwt, "decode", lambda *args, **kwargs: {"tenantId": "test"})


@pytest.fixture(scope="function")
def mock_version_rename(sw, monkeypatch):
    the_response = Response()
    the_response.status_code = 200

    with open(f"{PROJECT_ROOT}/tests/resources/data_entity/assets/entities.json", "r") as fh:
        data_entities = fh.read()
    example_model = Model(name="test", description="test")
    example_res = {"name": "new name"}
    the_response._content = example_res

    monkeypatch.setattr(requests, "patch", lambda *args, **kwargs: the_response)


@pytest.fixture(scope="function")
def mock_version_activate(sw, monkeypatch):
    the_response = Response()
    the_response.status_code = 200

    with open(f"{PROJECT_ROOT}/tests/resources/data_entity/assets/entities.json", "r") as fh:
        data_entities = fh.read()
    example_model = Model(name="test", description="test")
    example_res = {"version_id": 1}
    the_response._content = example_res

    monkeypatch.setattr(requests, "patch", lambda *args, **kwargs: the_response)


@pytest.fixture(scope="function")
def mock_version_requests(sw, monkeypatch):
    the_response = Response()
    the_response.status_code = 201
    with open(f"{PROJECT_ROOT}/tests/resources/data_entity/assets/entities.json", "r") as fh:
        data_entities = fh.read()
    data_response_response = Response()
    data_response_response._content = data_entities
    data_response_response.status_code = 200
    example_model = Model(name="test", description="test")
    example_version = Version(model_id=1, status="Active", name="test version")

    the_response._content = json.dumps(example_version.get_properties())
    monkeypatch.setattr(requests, "post", lambda *args, **kwargs: the_response)
    monkeypatch.setattr(sw.model, "get_by_id", lambda *args, **kwargs: example_model)
    monkeypatch.setattr(requests, "get", lambda *args, **kwargs: the_response)
    request_get_lambda = lambda url, **kwargs: data_response_response if "data_entities" in url else the_response
    monkeypatch.setattr(requests, "get", request_get_lambda)


def get_entities_fixture(path=f"{PROJECT_ROOT}/tests/resources/basic_schema.json"):
    with open(path, "r") as fh:
        schema = json.loads(fh.read())

    entities = [
        DataEntity(
            dimension_start_ts=m.get("dimension_start_ts", None),
            name=m["name"],
            type=m["type"],
            role=m["role"],
            feature_importance=None,
        )
        for m in schema
    ]
    return entities


def test_dataentity_creation_unit():

    entities = DataEntity(
        dimension_start_ts=None,
        name="barak",
        type=FeatureType.CATEGORICAL,
        role=DataEntityRole.LABEL,
        feature_importance=None,
    )

    assert entities.type == FeatureType.CATEGORICAL.value
    assert entities.role == DataEntityRole.LABEL.value


# source https://stackoverflow.com/questions/32815640/how-to-get-the-difference-between-two-dictionaries-in-python
def diff_dicts(a, b, missing=KeyError):
    """
    Find keys and values which differ from `a` to `b` as a dict.

    If a value differs from `a` to `b` then the value in the returned dict will
    be: `(a_value, b_value)`. If either is missing then the token from
    `missing` will be used instead.

    :param a: The from dict
    :param b: The to dict
    :param missing: A token used to indicate the dict did not include this key
    :return: A dict of keys to tuples with the matching value from a and b
    """
    return {key: (a.get(key, missing), b.get(key, missing)) for key in dict(set(a.items()) ^ set(b.items())).keys()}


def test_infer_dtypes(sw, mock_version_requests):
    df = pd.read_json(f"{PROJECT_ROOT}/tests/resources/internal_sdk/data_bool.json")
    expected_infer_path = f"{PROJECT_ROOT}/tests/resources/data_entity/expected_infer.json"
    with open(expected_infer_path) as f:
        expected_infer = json.load(f)

    dtypes_infered = infer_dtype(df)
    try:
        assert dtypes_infered == expected_infer
    except:
        diff = diff_dicts(expected_infer, dtypes_infered)
        print(diff)
        raise Exception("infered datatypes not same as schema")


def test_create_version(sw, mock_version_requests):
    # entities = get_entities_fixture()
    df = pd.read_json(f"{PROJECT_ROOT}/tests/resources/internal_sdk/data_bool.json")
    schema_path = f"{PROJECT_ROOT}/tests/resources/internal_sdk/basic_schema.json"
    with open(schema_path) as f:
        schema = json.load(f)

    roles = {}
    dtypes = {}
    for entity in schema:
        if entity["role"] != "feature":
            roles[entity["name"]] = entity["role"]
        dtypes[entity["name"]] = entity["type"]

    # Example of dtypes
    # dtypes = {
    #     "f0": "Numeric",
    #     "binary_bool_full": "Boolean",
    #     "binary_float_full": "Categorical",
    #     "ts": "Timestamp",
    #     "record_id": "Categorical",
    # }

    # Example of roles
    # roles = {
    #     "f0": "feature",
    #     "binary_bool_full": "feature",
    #     "binary_float_full": "feature",
    #     "ts": "time stamp",
    #     "record_id": "id",
    # }

    entities = sw.data_entity.summarise(
        data=df,
        entities_dtypes=dtypes,
        specific_roles=roles,
        default_role="feature",
        importance=False,
        importance_target_label=None,
        importance_sample=None,
        base_version=None,
    )
    version_external = Version(model_id=1, name="test version", data_entities=entities)
    model = sw.version.create(version_external, wait_until_complete=False)
    assert isinstance(model, Version)

    ## create a new one, but now with infer datatype
    entities = sw.data_entity.summarise(
        data=df,
        entities_dtypes=None,
        specific_roles=roles,
        default_role="feature",
        importance=False,
        importance_target_label=None,
        importance_sample=None,
        base_version=None,
    )
    version_external_2 = Version(model_id=1, name="test version with infer dtypes", data_entities=entities)
    model_2 = sw.version.create(version_external, wait_until_complete=False)
    assert isinstance(model_2, Version)


def test_create_base_version(sw, mock_version_requests):

    df = pd.read_json(f"{PROJECT_ROOT}/tests/resources/internal_sdk/data_bool.json")
    schema_path = f"{PROJECT_ROOT}/tests/resources/internal_sdk/basic_schema.json"
    with open(schema_path) as f:
        schema = json.load(f)
    df = df.rename(columns={"f31": "f31_change"})

    roles = {}
    dtypes = {}
    for entity in schema:
        if entity["name"] == "f31":
            entity["name"] = "f31_change"
        if entity["role"] != "feature":
            roles[entity["name"]] = entity["role"]
        dtypes[entity["name"]] = entity["type"]

    base_version = Version(id=1, status="Active")

    entities = sw.data_entity.summarise(
        data=df,
        entities_dtypes=dtypes,
        specific_roles=roles,
        default_role="feature",
        importance_target_label=None,
        importance_sample=None,
        base_version=base_version,
    )
    version_external = Version(model_id=1, name="test version", data_entities=entities)
    model = sw.version.create(version_external, wait_until_complete=False)
    assert isinstance(model, Version)
    print(model.get_properties())


def test_get_version(sw, mock_version_requests):
    version = sw.version.get_by_id(1)
    assert isinstance(version, Version)


def test_activate_version(sw, mock_version_activate):
    version = sw.version.activate(1)
    assert version.content["version_id"] == 1


def test_rename_version(sw, mock_version_rename):
    version = sw.version.update(1, {"name": "moshe"})
    assert version.content["name"] == "new name"


def test_get_version_data_entities(sw, mock_version_requests):
    data_entities = sw.version.get_data_entities(version_id=1)
    assert isinstance(data_entities, list)
    assert isinstance(data_entities[0], DataEntity)



""" This module implement Project model  """
import datetime

from superwise.models.base import BaseModel


class Project(BaseModel):
    """Project model class"""

    def __init__(
        self,
        id: int = None,
        name: str = None,
        description: str = None,
        created_at: datetime.datetime = None,
        created_by: str = None,
        **kwargs
    ):
        """
        ### Description:

        Constructor for Model class

        ### Args:

        `id`: id of project

        `name`: name of project

        `description`: description for the project

        `created_at`: date of creation

        created_by: user who created the project
        """
        self.id = id
        self.name = name
        self.description = description
        self.created_at = created_at
        self.created_by = created_by



""" This module implement infer dtypes functionality  """
import numpy as np
import pandas as pd


def infer_dtype(df):
    """
    ### Description:

     This function infer dtypes of a given df

    ### Args:

    `df`: a pandas df of baseline/data


    ### Return:

    dictionary <feature_name:dtype_string>
    """
    res = df.apply(_infer_entity_type)
    return res.to_dict()


def _infer_entity_type(feature):
    """
    ### Description:

     Infer an entity type (dtype)

    ### Args:

    `feature`:  pandas series of a feature


    ### Return:

    string represent the type of entity
    """

    """
    Infer an entity type (dtype)
    :param feature: feature pandas series
    """
    nunique = feature.nunique()
    values = feature.dropna()

    if nunique == 0:
        return "Unknown"
    if np.issubdtype(feature.dtype.base, np.bool_) or (
        (np.issubdtype(feature.dtype.base, np.number) and values.isin([1, 0]).all())
    ):
        return "Boolean"
    if np.issubdtype(feature.dtype.base, np.number):
        return "Numeric"
    if np.issubdtype(feature.dtype.base, np.datetime64):
        return "Timestamp"
    if np.issubdtype(pd.to_datetime(feature, errors="ignore").dtype.base, np.datetime64):
        return "Timestamp"
    if values.apply(np.isscalar).all():
        return "Categorical"
    return "Unknown"



""" This module implement validation errors model  """
from superwise.models.base import BaseModel
from superwise.utils.exceptions import *


class ValidationError(BaseModel):
    """validation error model class"""

    def __init__(self, http_status_code=None, http_error_reason=None, body=None):
        """
        ### Description:

        Constructor for ValidationError, this class used for handling validation errors

        ### Args:

         `http_status_code`: http status code from server

         `http_error_reason`: reason for error

         `body`: body of response object
        """
        self.http_status_code = http_status_code
        self.http_error_reason = http_error_reason
        self.details = []
        if isinstance(body, str):
            self.error = body
        else:
            detail = body.get("detail")
            if detail:
                self.error = "Input Validation error"
                if isinstance(detail, list):
                    for item in detail:
                        self.details.append(
                            {
                                "field": ".".join(item["loc"]),
                                "error": "{} in field {}".format(item["msg"], item["loc"][1]),
                            }
                        )
                else:
                    self.details = detail
        raise SuperwiseValidationException(self.get_properties())



"""Welcome to the superwise SDK reference guide!</br>
    This page is intended to help you better understand the capabilities of our SDK.</br>
    Here you will find all information about the SDK controllers, enums and models. </br> """
import json
import logging
import os
import pkgutil

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

from superwise.config import Config, DefaultValues
from superwise.utils.client import Client
from superwise.controller.model import ModelController
from superwise.utils.storage.internal_storage.internal_storage_factory import InternalStorageFactory
from superwise.controller.version import VersionController
from superwise.controller.role import RoleController
from superwise.controller.dataentity import DataEntityController
from superwise.controller.transaction import TransactionController
from superwise.controller.segment import SegmentController
from superwise.controller.notification import NotificationController
from superwise.controller.policy import PolicyController
from superwise.controller.project import ProjectController
from superwise.controller.dataset import DatasetController


class Superwise:
    """superwise class  main class for superwise package"""

    def __init__(
        self,
        client_id=None,
        secret=None,
        _rest_client=None,
        email=None,
        password=None,
        _auth_url=None,
        superwise_host=None,
    ):
        """
        ### Description:

        Constructor for superwise class

        ### Args:

        `client_id`: client access token

        `secret`: secret access token

        `email`: email of user (optional)

        `password`: password of user (optional)


        """

        self.logger = logger
        if superwise_host:
            Config.SUPERWISE_HOST = superwise_host
        if _auth_url:
            Config.AUTH_URL = _auth_url
        else:
            if Config.SUPERWISE_HOST.endswith(DefaultValues.SUPERWISE_DOMAIN.value):
                # env is saas
                Config.AUTH_URL = DefaultValues.AUTH_URL.value
            else:
                # env is on-prem
                Config.AUTH_URL = DefaultValues.MANAGED_AUTH_URL.value
        client_id = client_id or os.environ.get("SUPERWISE_CLIENT_ID")
        secret = secret or os.environ.get("SUPERWISE_SECRET")
        if email and password:
            self.logger.info("login using user and password")
        elif secret is None or client_id is None:
            raise Exception("secret or email/password are mendatory fields")
        api_host = Config.SUPERWISE_HOST
        if not _rest_client:
            _rest_client = Client(client_id, secret, api_host, email, password)
        self._set_config(_rest_client)
        internal_bucket = InternalStorageFactory.get_internal_bucket(_rest_client, Config.INTERNAL_BUCKET_SETTINGS)
        self.tenant_id = _rest_client.tenant_id
        self.model = ModelController(_rest_client, self)
        self.version = VersionController(_rest_client, self)
        self.data_entity = DataEntityController(_rest_client, self)
        self.transaction = TransactionController(_rest_client, self, internal_bucket)
        self.role = RoleController(_rest_client, self)
        self.notification = NotificationController(_rest_client, self)
        self.policy = PolicyController(_rest_client, self)
        self.project = ProjectController(_rest_client, self)
        self.segment = SegmentController(_rest_client, self)
        self.dataset = DatasetController(_rest_client, self, internal_bucket)

    @classmethod
    def _get_admin_settings(cls, superwise_web_client: Client):
        settings_resp = superwise_web_client.get(superwise_web_client.build_url("admin/v1/settings"))
        settings_resp.raise_for_status()

        settings = settings_resp.json()
        return settings

    @classmethod
    def _set_config(cls, superwise_web_client: Client):
        admin_settings = cls._get_admin_settings(superwise_web_client)
        Config.INTERNAL_BUCKET_SETTINGS = admin_settings.get("file_upload_settings")
